# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Shell completion formatters"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from functools import reduce
from os.path import basename

from .utils import args2sgpl
from .utils_2to3 import basestring, reduce_uu
from .utils_func import bifilter_unpack
from .utils_prog import cli_undecor


class Completion(object):
    mapping = {}

    def __init__(self, prog, opts_common, opts_main, opts_nonmain):
        self._prog = prog
        self._opts_common = opts_common
        self._opts_main = opts_main
        self._opts_nonmain = opts_nonmain

    def scripts_prologue(self):
        return ''

    def handle_script(self, command):
        raise RuntimeError('subclasses ought to override this')

    def scripts_epilogue(self, handles, aliases):
        return ''

    def __call__(self, commands):
        cmds, aliases = bifilter_unpack(
            lambda name, obj: not isinstance(obj, basestring),
            commands
        )
        handles, scripts = reduce_uu(
            lambda acc_handle, acc_script, cmd_name, cmd:
                (lambda handle, script: (
                    acc_handle + [(cmd_name, handle)],
                    acc_script + [script]
                ))(*self.handle_script(cmd)),
            cmds,
            ([], [])
        )
        scripts = [self.scripts_prologue(), ] + scripts
        scripts.append(self.scripts_epilogue(handles, aliases))
        return '\n\n'.join(scripts)

    @classmethod
    def deco(basecls, which):
        def deco(cls):
            basecls.mapping[which] = cls
            return cls
        return deco

    @classmethod
    def get_completion(cls, which, *args):
        return cls.mapping[which](*args)


@Completion.deco("bash")
class BashCompletion(Completion):
    # XXX include expert options?
    def __init__(self, prog, *args):
        prog = basename(prog)
        super(BashCompletion, self).__init__(prog, *args)
        self._name = cli_undecor(prog)

    @staticmethod
    def _namespaced_identifier(namespace, name=None):
        return '_'.join(filter(lambda x: x is not None, ('', namespace, name)))

    @staticmethod
    def _format_function(name, bodylines):
        bodylines = args2sgpl(bodylines)
        return ("{0}() {{\n\t{1}\n}}"
                .format(name, '\n\t'.join(bodylines).rstrip('\t')))

    @staticmethod
    def scripts_prologue():
        return """\
# bash completion start
# add me to ~/.profile persistently or eval (enquoted!) on-the-fly in bash"""

    def handle_script(self, cmd):
        clsname = cmd.__class__.__name__
        handle = self._namespaced_identifier(self._name, clsname)
        _, opts = cmd.parser_desc_opts()  # XXX not accurate without other opts
        main = """\
local opts="{0}"

[[ "$1" =~ -.* ]] && compgen -W "${{opts}}" -- $1"""\
        .format(
            '\n'.join(reduce(lambda a, b:
                             a + [' '.join(i + j for i in b[0] for j in (('', '=')
                                 if b[1].get('action', 'store')
                                     in ('callback', 'store')
                                 else ('', )
                             ))], opts, []))
        ).splitlines()

        handle = cli_undecor(handle)
        return handle, self._format_function(handle, main)

    def scripts_epilogue(self, handles, aliases):
        handle = self._namespaced_identifier(self._name)
        opts_common, opts_main, opts_nonmain = tuple(
            '\n'.join(reduce(lambda a, b:
                             a + [' '.join(i + j for i in b[0] for j in (('', '=')
                                 if b[1].get('action', 'store')
                                     in ('callback', 'store')
                                 else ('', )
                             ))], o, []))
            for o in (self._opts_common, self._opts_main, self._opts_nonmain)
        )
        alias_case = '    ' + '\n    '.join(
            '{0}) cur="{1}";;'.format(alias, to) for alias, to in aliases
        )
        # usage of self._name: see self._namespaced_identifier
        main = \
r"""local commands="{1}"
local opts_common="{2}"
local opts_main="{3}"
local opts_nonmain="{4}"

local cur fnc i=${{COMP_CWORD}}
while true; do
    test ${{i}} -eq 0 && break || let i-=1
    cur=${{COMP_WORDS[${{i}}]}}
    [[ "${{cur}}" =~ ^[-=].* ]] && continue
    # handle aliases
    case ${{cur}} in
{5}
    esac
    fnc=_{0}_${{cur/-/_}}
    declare -f ${{fnc}} >/dev/null \
     && COMPREPLY+=( $(${{fnc}} $2) ) \
     || continue
    [[ "$2" =~ ^-.* ]] \
     && COMPREPLY+=( $(compgen -W "${{opts_common}} ${{opts_nonmain}}" -- $2) )
    return
done

case "$2" in
-*) COMPREPLY=( $(compgen -W "${{opts_common}} ${{opts_main}}" -- $2) );;
*)  COMPREPLY=( $(compgen -W "${{commands}}" -- $2) );;
esac""" .format(
            self._name, '\n'.join(a for a, _ in (aliases + handles)),
            opts_common, opts_main, opts_nonmain, alias_case
        ).splitlines()
        epilogue = "complete -o default -F {0} {1}".format(handle, self._prog)
        return '\n\n'.join([self._format_function(handle, main), epilogue])



# -*- coding: UTF-8 -*-
# Copyright 2016 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....filters._2pcscmd import verbose_ec_test, verbose_inform
from ....utils_xslt import NL, translate_lower, xslt_is_member

cib2pcscmd_datespec = (
    'hours',
    'monthdays',
    'weekdays',
    'yearsdays',
    'months',
    'weeks',
    'years',
    'weekyears',
    'moon',
)

cib2pcscmd = ('''\
    <xsl:choose>
        <!-- plain prefers/avoids (XXX could be ordered for prettiness) -->
        <xsl:when test="@score">
            <xsl:variable name="Relationship">
                <xsl:choose>
                    <xsl:when test="@score = 'INFINITY'
                                    or
                                    @score = '+INFINITY'
                                    or
                                    @score &gt;= 0">
                        <xsl:value-of select="'prefers'"/>
                    </xsl:when>
                    <xsl:otherwise>
                        <xsl:value-of select="'avoids'"/>
                    </xsl:otherwise>
                </xsl:choose>
            </xsl:variable>
            <xsl:value-of select="concat($pcscmd_pcs, 'constraint location',
                                        ' ', @rsc,
                                        ' ', $Relationship, ' ', @node)"/>
            <xsl:if test="not(contains(@score, 'INFINITY'))">
                <xsl:value-of select="concat('=', @score)"/>
            </xsl:if>
            <xsl:value-of select="'%(NL)s'"/>
        </xsl:when>

        <!-- rule-based location constraint -->
        <xsl:when test="rule">
            <xsl:variable name="Resource" select="@rsc"/>
            <xsl:variable name="ConstraintId" select="@id"/>
            <xsl:for-each select="rule[@id-ref]">
                <!-- XXX: could eventually be unfolded in-place if found -->
                <xsl:message
                >WARNING: skipping location rule provided via reference</xsl:message>
            </xsl:for-each>
            <xsl:for-each select="rule[not(@id-ref)]">
''' + (
                verbose_inform('"new rule/location constraint: ", @id, "/", $ConstraintId')
) + '''
                <xsl:variable name="BooleanOp" select="@boolean-op"/>
                <xsl:choose>
                    <xsl:when test="position() = 1">
                        <xsl:value-of select="concat(
                                                $pcscmd_pcs, 'constraint location',
                                                ' ', $Resource,
                                                ' ', 'rule',
                                                ' ', 'id=', @id,
                                                ' ', 'constraint-id=', $ConstraintId
                                            )"/>
                    </xsl:when>
                    <xsl:otherwise>
                        <!-- subsequent per-constraint rules;
                            see https://bugzilla.redhat.com/1223404 -->
                        <xsl:value-of select="concat(
                                                $pcscmd_pcs, 'constraint rule add',
                                                ' ', $ConstraintId,
                                                ' ', 'id=', @id
                                            )"/>
                    </xsl:otherwise>
                </xsl:choose>
                <xsl:if test="
''' + (
                xslt_is_member(translate_lower('@role'),
                               ('master', 'slave'))
) + '''">
                    <xsl:value-of select="concat(' ',
''' + (
                        translate_lower('@role')
) + ''')"/>
                </xsl:if>
                <xsl:for-each select="@*[
''' + (
                    xslt_is_member('name()', ('score', 'score-attribute'))
) + ''']">
                    <xsl:value-of select="concat(' ', name(), '=', .)"/>
                </xsl:for-each>

                <xsl:if test="rule">
                    <xsl:message
                    >WARNING: cannot handle nested rules (yet)</xsl:message>
                </xsl:if>

                <xsl:for-each select="descendant::expression|descendant::date_expression">
                    <xsl:if test="count(preceding-sibling::expression|preceding-sibling::date_expression) &gt; 0">
                        <xsl:value-of select="concat(' ', $BooleanOp)"/>
                    </xsl:if>
                    <xsl:choose>
                        <xsl:when test="name() = 'expression'">
                            <xsl:choose>
                                <xsl:when test="
''' + (
                                xslt_is_member('@operation', ('defined', 'not_defined'))
) + '''">
                                    <xsl:value-of select='concat(" ", @operation,
                                                                " &apos;", @attribute, "&apos;")'/>
                                </xsl:when>
                                <xsl:otherwise>
                                    <xsl:value-of select='concat(" &apos;", @attribute, "&apos;",
                                                                " ", @operation)'/>
                                    <xsl:if test="
''' + (
                                    xslt_is_member('@type', ('string', 'number', 'version'))
) + '''">
                                        <xsl:value-of select="concat(' ', @type)"/>
                                    </xsl:if>
                                    <xsl:value-of select='concat(" &apos;", @value, "&apos;")'/>
                                </xsl:otherwise>
                            </xsl:choose>
                        </xsl:when>
                        <xsl:when test="name() = 'date_expression'">
                            <xsl:choose>
                                <xsl:when test="@operation = 'in_range'">
                                    <xsl:value-of select="' date in_range '"/>
                                    <xsl:choose>
                                        <xsl:when test="@end">
                                            <!-- see https://bugzilla.redhat.com/1182358 -->
                                            <xsl:if test="@start">
                                                <xsl:value-of select="@start"/>
                                            </xsl:if>
                                            <xsl:value-of select="concat(' to ', @end)"/>
                                        </xsl:when>
                                        <xsl:when test="duration">
                                            <xsl:value-of select="concat(@start, ' to ')"/>
                                            <xsl:for-each select="duration/@*[
''' + (
                                            xslt_is_member('name()', cib2pcscmd_datespec)
) + ''']">
                                                <xsl:value-of select="concat(' ', name(), '=', .)"/>
                                            </xsl:for-each>
                                        </xsl:when>
                                    </xsl:choose>
                                </xsl:when>
                                <xsl:when test="@operation = 'gt'">
                                    <xsl:value-of select="concat(' date gt ', @start)"/>
                                </xsl:when>
                                <xsl:when test="@operation = 'lt'">
                                    <xsl:value-of select="concat(' date lt ', @end)"/>
                                </xsl:when>
                                <xsl:when test="@operation = 'date_spec'">
                                    <xsl:value-of select="' date-spec '"/>
                                        <xsl:for-each select="date_spec/@*[
''' + (
                                        xslt_is_member('name()', cib2pcscmd_datespec)
) + ''']">
                                            <xsl:value-of select="concat(' ', name(), '=', .)"/>
                                        </xsl:for-each>
                                </xsl:when>
                            </xsl:choose>
                        </xsl:when>
                    </xsl:choose>
                </xsl:for-each>
                <xsl:value-of select="'%(NL)s'"/>
''' + (
                verbose_ec_test
) + '''
            </xsl:for-each>
        </xsl:when>
    </xsl:choose>

''') % dict(
    NL=NL,
)



# -*- coding: UTF-8 -*-
# Copyright 2014 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

ccs2ccs_pcmk = '''\
    <fencedevices>
        <fencedevice agent="fence_pcmk"
                     name="pcmk-redirect"/>
    </fencedevices>
'''



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....utils_cib import ResourceSpec, rg2hb_xsl


ccsflat2cibprelude = '''\
    <!--
        named ~ named (but only since resource-agents[named])
     -->
    <xsl:when test="name() = 'named' and $pcscmd_named_qual">
        <xsl:message terminate="yes">
            <xsl:value-of select="concat('Cannot convert named when the',
                                         ' target distro does not qualify: ',
                                         $pcscmd_named_qual)"/>
        </xsl:message>
    </xsl:when>
    <xsl:when test="name() = 'named'">
''' + (
        ResourceSpec('ocf:heartbeat:named').xsl_attrs
) + '''
        <!-- INSTANCE_ATTRIBUTES -->
        <instance_attributes id="{concat($Prefix, '-ATTRS')}">
''' + (
            rg2hb_xsl('named_config', 'config_file')
            +
            rg2hb_xsl('named_options')
) + '''\
        </instance_attributes>

        <!-- OPERATIONS -->
        <operations>
''' + (
            rg2hb_xsl('stop', 'shutdown_wait', op=True)
) + '''\
        </operations>
    </xsl:when>
'''

###

from ....filters.ccs_artefacts import artefact_cond_ra

ccs_artefacts = ''.join((
    artefact_cond_ra('@config_file',
                     kind='A', desc='configuration file'),
))



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""ccs2pcscmd{,-flatiron,-needle} commands"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..command import Command, CommandAlias
try:
    from ..defaults import SHELL_BASHLIKE, SHELL_POSIX
except ImportError:
    SHELL_BASHLIKE, SHELL_POSIX = '', ''
from ..facts import cluster_pcs_flatiron
from ..filter import XMLFilter
from ..protocol import protocols
from ..utils_cman import PATH_CLUSTERCONF
from ._chains_pcs import ccsflat2pcscmd_chain_exec, output_set_exec


@Command.deco(('cmd-annotate',
                                  ('stringiter-combine3',
                                      ('cmd-wrap'))),
              ('ccs2ccsflat',
                  ('ccs-disable-rg',
                      ('ccs2ccs-pcmk',
                          ('ccs-version-bump',
                              ('ccspcmk2pcscmd',
                                  ('stringiter-combine3'  # , ('cmd-wrap' ...
                                   ))))),
                  (ccsflat2pcscmd_chain_exec(
                                  ('stringiter-combine3'  # , ('cmd-wrap' ...
                                   )))))
def ccs2pcscmd_flatiron(cmd_ctxt,
                        input=PATH_CLUSTERCONF,
                        output="-",
                        force=False,
                        noauth=False,
                        silent=False,
                        tmp_cib="{cib2pcscmd.defs[pcscmd_tmpcib]}",
                        dry_run=False,
                        enable=False,
                        start_wait="{ccspcmk2pcscmd.defs[pcscmd_start_wait]}",
                        noguidance=False,
                        set_exec=False,
                        text_width='0',
                        _common=XMLFilter.command_common):
    """(CMAN,rgmanager) cluster cfg. -> equivalent in pcs commands

    Options:
        input       input (CMAN,rgmanager) cluster configuration file
        output      pcs commands to reinstate the cluster per the inputs
        force       may the force be with emitted pcs commands
        noauth      skip authentication step (OK if already set up)
        silent      do not track the progress along the steps execution (echoes)
        tmp_cib     file to accumulate the changes (empty ~ direct push, avoid!)
        dry_run     omit intrusive commands (TMP_CIB reset if empty)
        enable      enable cluster infrastructure services (autostart on reboot)
        start_wait  fixed seconds to give cluster to come up initially
        noguidance  omit extraneous guiding
        set_exec    make the output file executable (not recommended)
        text_width  for commands rewrapping (0/-1/neg. ~ auto/disable/hi-limit)
    """
    cmd_ctxt['pcscmd_force'] = force
    cmd_ctxt['pcscmd_noauth'] = noauth
    cmd_ctxt['pcscmd_verbose'] = not(silent)
    cmd_ctxt['pcscmd_tmpcib'] = tmp_cib
    cmd_ctxt['pcscmd_dryrun'] = dry_run
    cmd_ctxt['pcscmd_enable'] = enable
    cmd_ctxt['pcscmd_start_wait'] = start_wait
    cmd_ctxt['pcscmd_noguidance'] = noguidance
    cmd_ctxt['text_width'] = text_width

    # possible use of process substitution (https://bugzilla.redhat.com/1381531)
    cmd_ctxt['annotate_shell'] = (SHELL_POSIX if dry_run or noguidance
                                  else SHELL_BASHLIKE)
    # XXX possibility to disable cib-meld-templates

    void_proto = protocols.plugins['void'].ensure_proto
    file_proto = protocols.plugins['file'].ensure_proto
    yield (
        (
            void_proto(),
            (
                                                (
                                                    file_proto(output),
                                                ),
            ),
        ), (
            file_proto(input),
            # already tracked
            #(
            #    (
            #        (
            #            (
            #                (
            #                                    (
            #                                        file_proto(output),
            #                                    ),
            #                ),
            #            ),
            #        ),
            #    ),
            #    # already tracked
            #    #ccsflat2pcscmd_output(
            #    #                                (
            #    #                                    file_proto(output),
            #    #                                ),
            #    #),
            #),
        ),
    )
    # post-processing (make resulting file optionally executable)
    if set_exec:
        output_set_exec(cmd_ctxt, 'cmd-wrap')


@Command.deco(('cmd-annotate',
                              ('stringiter-combine4',
                                  ('cmd-wrap'))),
              ('ccs2ccsflat',
                  ('ccs-propagate-cman',
                      ('ccs2needlexml',
                          ('needlexml2pcscmd',
                              ('stringiter-combine4'  # , ('cmd-wrap' ...
                               )),
                          ('needleqdevicexml2pcscmd',
                              ('stringiter-combine4'  # , ('cmd-wrap' ...
                               )))),
                  (ccsflat2pcscmd_chain_exec(
                              ('stringiter-combine4'  # , ('cmd-wrap' ...
                               )))))
def ccs2pcscmd_needle(cmd_ctxt,
                      input=PATH_CLUSTERCONF,
                      output="-",
                      force=False,
                      noauth=False,
                      silent=False,
                      tmp_cib="{cib2pcscmd.defs[pcscmd_tmpcib]}",
                      dry_run=False,
                      enable=False,
                      start_wait="{needlexml2pcscmd.defs[pcscmd_start_wait]}",
                      noguidance=False,
                      set_exec=False,
                      text_width='0',
                      _common=XMLFilter.command_common):
    """(CMAN,rgmanager) cluster cfg. -> equivalent in pcs commands

    Options:
        input       input (CMAN,rgmanager) cluster configuration file
        output      pcs commands to reinstate the cluster per the inputs
        force       may the force be with emitted pcs commands
        noauth      skip authentication step (OK if already set up)
        silent      do not track the progress along the steps execution (echoes)
        tmp_cib     file to accumulate the changes (empty ~ direct push, avoid!)
        dry_run     omit intrusive commands (TMP_CIB reset if empty)
        enable      enable cluster infrastructure services (autostart on reboot)
        start_wait  fixed seconds to give cluster to come up initially
        noguidance  omit extraneous guiding
        set_exec    make the output file executable (not recommended)
        text_width  for commands rewrapping (0/-1/neg. ~ auto/disable/hi-limit)
    """
    cmd_ctxt['pcscmd_force'] = force
    cmd_ctxt['pcscmd_noauth'] = noauth
    cmd_ctxt['pcscmd_verbose'] = not(silent)
    cmd_ctxt['pcscmd_tmpcib'] = tmp_cib
    cmd_ctxt['pcscmd_dryrun'] = dry_run
    cmd_ctxt['pcscmd_enable'] = enable
    cmd_ctxt['pcscmd_start_wait'] = start_wait
    cmd_ctxt['pcscmd_noguidance'] = noguidance
    cmd_ctxt['text_width'] = text_width

    # possible use of process substitution (https://bugzilla.redhat.com/1381531)
    cmd_ctxt['annotate_shell'] = (SHELL_POSIX if dry_run or noguidance
                                  else SHELL_BASHLIKE)
    # XXX possibility to disable cib-meld-templates

    void_proto = protocols.plugins['void'].ensure_proto
    file_proto = protocols.plugins['file'].ensure_proto
    yield (
        (
            void_proto(),
            (
                                                (
                                                    file_proto(output),
                                                ),
            ),
        ), (
            file_proto(input),
            #(
            #    (
            #        (
            #                                (
            #                                    (
            #                                        file_proto(output),
            #                                    ),
            #                                ),
            #            # already tracked
            #            #                    (
            #            #                        (
            #            #                            file_proto(output),
            #            #                        ),
            #            #                    ),
            #        ),
            #    ),
            #    # already tracked
            #    #ccsflat2pcscmd_output(
            #    #                                (
            #    #                                    file_proto(output),
            #    #                                ),
            #    #),
            #),
        ),
    )
    # post-processing (make resulting file optionally executable)
    if set_exec:
        output_set_exec(cmd_ctxt, 'cmd-wrap')


@CommandAlias.deco
def ccs2pcscmd(cmds, *sys_id):
    # cluster_pcs_needle assumed unless "cluster_pcs_flatiron"
    return (ccs2pcscmd_flatiron if cluster_pcs_flatiron(*sys_id) else
            ccs2pcscmd_needle)



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""pcs_resource_deps command"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..command import Command
from ..filter import XMLFilter
from ..protocol import protocols

@Command.deco(('pcs-resource-deps',
                   ('stringlist2stringset',
                        ('pkgs2distropkgs',
                            ('pkgs2installcmd')))))
def pcs_rsc_deps(cmd_ctxt,
                 input="/var/lib/pacemaker/cib/cib.xml",
                 output="pcs-resource-deps-{pcs-resource-deps.in.hash}.conf",
                 install_cmd=False,
                 _common=XMLFilter.command_common):
    """Output deps of resources contained in Pacemaker cfg.

    Options:
        input        input Pacemaker configuration file
        output       output file with collected dependencies
        install_cmd  whether to output full install command instead
    """
    #if not install_cmd:
    #    cmd_ctxt['filter_noop'].append('pkgs2installcmd')

    file_proto = protocols.plugins['file'].ensure_proto
    return (
        file_proto(input),
        (
            (
                (
                    file_proto(output),
                ),
            ),
        ),
    )



# -*- coding: UTF-8 -*-
# Copyright 2014 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Base exception classes and exit code definitions"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"


# name the exitcodes
ecodes = 'SUCCESS', 'FAILURE'
EC = type('EC', (), dict((n, v) for v, n
                         in enumerate('EXIT_' + i for i in ecodes)))


class ClufterError(Exception):
    def __init__(self, ctx_self, msg, *args):
        self.ctx_self = ctx_self
        self.msg = msg
        self.args = args

    def __str__(self):
        ret = ''
        if self.ctx_self:
            ret = getattr(self.ctx_self, '__name__',
                          self.ctx_self.__class__.__name__)
            ret += ': '
        return ret + self.msg.format(*self.args)


class ClufterPlainError(ClufterError):
    def __init__(self, msg, *args):
        super(ClufterPlainError, self).__init__(None, msg, *args)



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""cmd-wrap filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..filter import Filter
from ..formats.command import command, ismetaword
from ..utils import head_tail
from ..utils_2to3 import bytes_enc, str_enc, xrange
from ..utils_func import add_item
from ..utils_prog import FancyOutput

from collections import MutableMapping, defaultdict
from logging import getLogger
from os import getenv, isatty
from sys import maxsize
from textwrap import TextWrapper

n, len_n = FancyOutput.normalized, FancyOutput.len_normalized
log = getLogger(__name__)

pcs_commands_hierarchy = {
    'cluster': {
        'auth': None,
        'setup': None,
        'start': None,
        'stop': None,
        'kill': None,
        'enable': None,
        'disable': None,
        # v-- deprecated (in favor of node {add,remove}-remote)
        'remote-node': {
            'add': None,
            'remove': None,
        },
        # ^-- deprecated
        'status': None,
        'pcsd-status': None,
        'sync': None,
        'cib': None,
        'cib-push': None,
        'cib-upgrade': None,
        'edit': None,
        'node': {
            'add': None,
            'remove': None,
            'add-remote': None,
            'remove-remote': None,
            'add-guest': None,
            'remove-guest': None,
            'clear': None,
        },
        'uidgid': {
            None: None,
            'add': None,
            'rm': None,
        },
        'corosync': None,
        'reload': {
            'corosync': None,
        },
        'destroy': None,
        'verify': None,
        'report': None,
    },
    'resource': {
        None: None,
        'show': None,
        'list': None,
        'describe': None,
        'create': None,
        'delete': None,
        'enable': None,
        'disable': None,
        'restart': None,
        'debug-start': None,
        'debug-stop': None,
        'debug-promote': None,
        'debug-demote': None,
        'debug-monitor': None,
        'move': None,
        'ban': None,
        'clear': None,
        'standards': None,
        'providers': None,
        'agent': None,
        'update': None,
        'op': {
            'add': None,
            'remove': None,
            'defaults': None,
        },
        'meta': None,
        'group': {
            'add': None,
            'remove': None,
        },
        'ungroup': None,
        'clone': None,
        'unclone': None,
        'master': None,
        'bundle': {
            'create': None,
            'update': None,
        },
        'manage': None,
        'unmanage': None,
        'defaults': None,
        'cleanup': None,
        'failcount': {
            'show': None,
            'reset': None,
        },
        'relocate': {
            'dry-run': None,
            'run': None,
            'show': None,
            'clear': None,
        },
        'utilization': None,
    },
    'stonith': {
        None: None,
        'show': None,
        'list': None,
        'describe': None,
        'create': None,
        'update': None,
        'delete': None,
        'cleanup': None,
        'level': {
            None: None,
            'add': None,
            'remove': None,
            'clear': None,
            'verify': None,
        },
        'fence': None,
        'confirm': None,
        'sbd': {
            'enable': None,
            'disable': None,
            'device': {
                'setup': None,
                'message': None,
            },
            'status': None,
            'config': None,
        },
    },
    'constraint': {
        None: None,
        'list': None,
        'show': None,
        'location': {
            None: None,
            'show': None,
            'remove': None,
        },
        'order': {
            None: None,
            'show': None,
            'set': None,
            'remove': None,
        },
        'colocation': {
            None: None,
            'show': None,
            'add': None,
            'set': None,
            'remove': None,
        },
        'ticket': {
            None: None,
            'show': None,
            'add': None,
            'set': None,
        },
        'remove': None,
        'ref': None,
        'rule': {
            'add': None,
            'remove': None,
        },
    },
    'property': {
        None: None,
        'list': None,
        'show': None,
        'set': None,
        'unset': None,
    },
    'acl': {
        None: None,
        'show': None,
        'enable': None,
        'disable': None,
        'role': {
            None: None,
            'create': None,
            'delete': None,
            'assign': None,
            'unassign': None,
        },
        'user': {
            None: None,
            'create': None,
            'delete': None,
        },
        'group': {
            None: None,
            'create': None,
            'delete': None,
        },
        'permission': {
            None: None,
            'add': None,
            'delete': None,
        },
    },
    'qdevice': {
        'status': None,
        'setup': {
            'model': None,
        },
        'destroy': None,
        'start': None,
        'stop': None,
        'kill': None,
        'enable': None,
        'disable': None,
    },
    'quorum': {
        None: None,
        'config': None,
        'status': None,
        'device': {
            'add': None,
            'remove': None,
            'status': None,
            'update': None,
        },
        'expected-votes': None,
        'unblock': None,
        'update': None,
    },
    'booth': {
        'setup': None,
        'destroy': None,
        'ticket': {
            'add': None,
            'remove': None,
            'grant': None,
        },
        'config': None,
        'create': {
            'ip': None,
        },
        'status': None,
        'pull': None,
        'sync': None,
        'enable': None,
        'disable': None,
        'start': None,
        'stop': None,
    },
    'status': {
        None: None,
        'resources': None,
        'groups': None,
        'cluster': None,
        'corosync': None,
        'quorum': None,
        'qdevice': None,
        'nodes': {
            None: None,
            'config': None,
            'corosync': None,
            'both': None,
        },
        'pcsd': None,
        'xml': None,
    },
    'config': {
        None: None,
        'show': None,
        'backup': None,
        'restore': None,
        'checkpoint': {
            None: None,
            'view': None,
            'restore': None,
        },
        'import-cman': None,
        'export': {
            'pcs-commands': None,
            'pcs-commands-verbose': None,
        },
    },
    'pcsd': {
        'certkey': None,
        'sync-certificates': None,
        'clear-auth': None,
    },
    'node': {
        'maintenance': None,
        'unmaintenance': None,
        'standby': None,
        'unstandby': None,
        'utilization': None,
    },
    'alert': {
        None: None,
        'config': None,
        'show': None,
        'create': None,
        'remove': None,
        'recipient': {
            'add': None,
            'update': None,
            'remove': None,
        },
    },
}


def cmd_args_colorizer_pcs(in_seq, color_map, initial=False,
                           level=pcs_commands_hierarchy):
    output = in_seq
    if len(in_seq) == 1 and initial and in_seq[0] == 'pcs':
        output = (in_seq[0].join((color_map['pcs'], color_map['restore'])), )
    else:
        output = []
        for i in add_item(in_seq, ''):
            if isinstance(level, MutableMapping):
                if i in level:
                    output.append(i)
                    level = level[i]
                    if level is None:
                        level = True
                    elif not isinstance(level, MutableMapping):
                        level = None  # XXX for the time being
                    continue  # need to yield new 'i' even if level is True
                level = None in level
            if level is True:
                last = len(output) - 1
                output = [ii.join((
                    color_map['subcmd'] if ee == 0 else '',
                    color_map['restore'] if ee == last else ''
                )) for (ee, ii) in enumerate(output)]
                level = None
            if i:
                output.append(i)
    return output


def cmd_args_cutter(itemgroups, color_map):
    if not itemgroups:
        return itemgroups

    ret, acc = [], []
    cmd = itemgroups[0][0] if itemgroups[0] else ""

    for e, i in enumerate(itemgroups):
        if len(i) and (not(i[0].startswith('-')) or i[0] == '-'):
            js = [f for f, j in enumerate(i) if ismetaword(j)]
            if js:
                js.extend((len(i), ) * 2)
                this_joint = 0
                for next_joint in js:
                    ret.extend(map(
                        lambda l: tuple(ii.join((
                            color_map['metaword'] if ismetaword(ii) else '',
                            color_map['restore'] if ismetaword(ii) else '',
                        )) for ee, ii in enumerate(l)),
                        filter(bool, (tuple(acc), ))
                    ))
                    acc = list(i[this_joint:next_joint])
                    this_joint = next_joint
            elif cmd.endswith('pcs'):
                pos = -1
                end = len(i)
                # separate and possibly colorize "pcs"
                if not e and i[0] == 'pcs':
                    ret.append(cmd_args_colorizer_pcs(i[0:1], color_map,
                                                      initial=True))
                    pos += 1
                while pos + 1 < end:
                    pos += 1
                    # try to cut into "firm groups"
                    if pos <= end - 4:
                        if i[pos:pos + 2] in (("resource", "create"),
                                              ("stonith", "create")):
                            # "resource/stonith create X Y" firm group
                            ret.extend(map(
                                lambda x: cmd_args_colorizer_pcs(x, color_map),
                                filter(bool, (tuple(acc), tuple(i[pos:pos + 4])))
                            ))
                            pos += 4
                            acc = list(i[pos:pos])
                            pos += len(acc) - 1
                            continue
                    if pos <= end - 3:
                        if i[pos:pos + 2] in (('property', 'set'),
                                              ('property', 'unset')):
                            # "property set/unset non-option [non-option...]"
                            ret.extend(map(
                                lambda x: cmd_args_colorizer_pcs(x, color_map),
                                filter(bool, (tuple(acc), ))
                            ))
                            acc = list(i[pos:pos + 2])
                            pos += len(acc) - 1
                            continue
                    if pos <= end - 2:
                        if i[pos] in (
                            "op", "meta",   # resource create/update
                            "id", "xpath",  # acl (role create|permission add)
                            # bundle feature
                            "container", "network", "port-map", "storage-map",
                                ):
                            ret.extend(map(
                                lambda x: cmd_args_colorizer_pcs(x, color_map),
                                filter(bool, (tuple(acc), ))
                            ))
                            acc = list(i[pos:pos + 1])
                            pos += len(acc) - 1
                            continue
                    # TBD
                    acc.append(i[pos])
                ret.append(cmd_args_colorizer_pcs(acc, color_map))
                acc = []
            else:
                ret.extend((ii, ) for ii in i)
        elif cmd.endswith('pcs'):
            if len(i) == 2:
                if i[0] == '--wait':
                    # --wait NUM has to be reverted into --wait=NUM due to the
                    # way pcs parser is "extended"
                    i = ( '='.join(i), )
                elif i[0] == '-f':
                    i = (ii.join((
                        color_map['file'] if ee == 0 else '',
                        color_map['restore'] if ee == 1 else ''
                    )) for ee, ii in enumerate(i))
            ret.append(tuple(i))
        else:
            ret.append(i)
    return ret


@Filter.deco('string-iter', 'string-iter')
def cmd_wrap(flt_ctxt, in_obj):
    """Try to apply a bit smarter wrapping on lines carrying shell commands

    Smarter means:
    - standard textwrap module logic applied on comments splitting; otherwise:
    - do not delimit option from its argument
    - when line is broken vertically, append backslash for the continuation
      and indent subsequent lines for visual clarity
    - and as a side-effect: normalize whitespace occurrences

    Width used for rewrapping is based on three factors in precedence order:
    - text_width value inside the filter context (`flt_ctxt`)
      - 0 ~ fall-through to the value per the next item
      - -1 ~ apply no wrapping at all (mimicked by implying huge text_width)
      - positive ~ hard-limit the width (no smartness involved)
      - negative ~ high-limit the width, apply the inverse value only when not
                   exceeding the value per the next item
    - COLUMNS environment variable, if defined and possesses integer value
    - hard-coded default of 72
    If absolute value at this point is lower than 20, fallback to 20.
    """
    proto, rest = head_tail(flt_ctxt.get('io_decl'))
    try:
        term = proto == 'file' and isatty(rest[0].fileno())
    except AttributeError:
        term = False
    color = flt_ctxt.get('color')
    color = color or (term and color is not False)
    tw, tw_system = 0, 0
    try:
        tw_system = int(getenv('COLUMNS')) if term else tw_system
    except TypeError:
        pass
    try:
        tw = int(flt_ctxt.get('text_width'))
    except TypeError:
        pass
    if tw > 0:
        pass
    elif tw < -1:
        tw = -tw if not tw_system or -tw < tw_system else tw_system
    elif tw == 0 and term:
        tw = tw_system
    else:
        tw = maxsize >> 1  # one order of magnitude less to avoid overflows
    if tw < 20:  # watch out for deliberate lower limit
        tw = 20 if tw else 72
        log.info('Text width fallback: {0}'.format(tw))
    cw = TextWrapper(width=tw, subsequent_indent='# ',  # wrapper for comments
                     break_on_hyphens=False)            # <-- for PY3
    color_map = (dict(((k, FancyOutput.get_color(
                               FancyOutput.table.get('pcscmd_' + k, '')))
                       for k in ('comment', 'file', 'metaword', 'pcs', 'subcmd')),
                        restore=FancyOutput.colors['restore'])
                if color else defaultdict(lambda: ''))

    ret, continuation = [], []
    for line in (str_enc(l, 'utf-8')
                 for l in (in_obj('bytestringiter', protect_safe=True))):
        if line.lstrip().startswith('#'):
            lines = cw.wrap(line)
            last = len(lines) - 1
            ret.extend(l.join((
                color_map['comment'] if e == 0 else '',
                color_map['restore'] if e == last else ''
            )) for e, l in enumerate(lines))
            continue
        # rough overapproximation of what is indeed a line continuation
        if line.endswith('\\') and not line.endswith('\\\\'):
            if '#' not in line:
                continuation.append(line[:-1])
                continue
            line += '\\'  # XXX
        line = ' '.join(continuation) + line
        continuation = []
        linecnt, rline, remains = -1, [], tw - 2  # ' \'
        itemgroups = cmd_args_cutter(command('bytestring', line)('separated'),
                                     color_map=color_map)
        itemgroups.reverse()
        while itemgroups:
            itemgroup = list(itemgroups.pop())
            itemgroup.reverse()
            while itemgroup:
                curlen = 0
                line = [itemgroup.pop()]
                curlen += len_n(line[-1])
                # best match fill
                while itemgroup \
                        and remains - (curlen + 1 + len_n(itemgroup[-1])) >= 0:
                    line.append(itemgroup.pop())
                    curlen += 1 + len_n(line[-1])
                # compensate for ' \' tail not necessary if very last item fits
                if not itemgroups and len(itemgroup) == 1 \
                        and len_n(itemgroup[-1]) == 1:
                    line.append(itemgroup.pop())
                    curlen += 1 + len_n(line[-1])
                # merge previous group to the current one if it fits the length
                if (rline
                    and not itemgroup
                    and remains - (curlen + 1 + len_n(' '.join(rline))) >= 0
                    and (not itemgroups
                         or not itemgroups[-1]
                         or not itemgroups[-1][0]
                         or not n(itemgroups[-1][0])[0] == '-'
                         or not ismetaword(n(line[0]))
                         or len(itemgroups) == 1 or not itemgroups[-2]
                         or ismetaword(n(itemgroups[-2][0])))):
                    line = rline + line
                    rline = []
                    linecnt -= 1
                # second pass optionally handles the terminal propagation
                for i in xrange(2):
                    if rline:
                        tail = ' \\' if rline is not line else ''
                        rline = ' '.join(rline)
                        if not linecnt:
                            ret.append(rline + tail)
                            remains -= 2  # initial indent
                        else:
                            ret.append('  ' + rline + tail)
                    linecnt += 1
                    rline = line
                    if itemgroups or itemgroup:
                        break
    return ('bytestringiter', (bytes_enc(l, 'utf-8') for l in ret))



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""ccs2ccsflat filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from logging import getLogger
from os.path import split as path_split
from subprocess import PIPE
from sys import stdin

from ..filter import Filter, FilterError
from ..utils_2to3 import basestring, str_enc
from ..utils_prog import OneoffWrappedStdinPopen, dirname_x, which

try:
    from ..defaults import CCS_FLATTEN
except ImportError:
    CCS_FLATTEN = 'ccs_flatten'

log = getLogger(__name__)


@Filter.deco('ccs', 'ccs-flat')
def ccs2ccsflat(flt_ctxt, in_obj):
    self = flt_ctxt.ctxt_wrapped

    ccsf_dirname, ccsf_basename = path_split(CCS_FLATTEN)
    # order of priority when searching for the binary (descending order):
    # - same as the root of the package (debugging purposes)
    # - directories-part as per specification in setup.cfg
    # - PATH env variable (if defined)
    ccs_flatten = which(ccsf_basename, dirname_x(__file__, 2), ccsf_dirname, '')
    if not ccs_flatten:
        raise FilterError(self, "ccs_flatten binary seems unavailable")

    # XXX currently ccs_flatten does not handle stdin (tempfile.mkstemp?)
    # XXX conversion is not idempotent, should prevent using ccs-flat as input
    #     (specifically, explicit ordering will get borken in subsequent round)
    in_file = in_obj('file')
    if not isinstance(in_file, basestring):
        # convert '-'/'@0' already converted to fileobj back to '-'
        if in_file is not stdin:
            raise RuntimeError("Only stdin ('-') supported")
        in_file = '-'
    command = [ccs_flatten, in_file]
    log.info("running `{0}'".format(' '.join(command)))
    try:
        proc = OneoffWrappedStdinPopen(command, stdout=PIPE, stderr=PIPE)
    except OSError:
        raise FilterError(self, "error running ccs_flatten binary")
    out, err = proc.communicate()
    if proc.returncode != 0 or (not out and err):
        raise FilterError(self, "ccs_flatten {0}\n\t{1}",
                          ("exit code: {0}" if proc.returncode > 0 else
                           "killing signal: {0}").format(abs(proc.returncode)),
                           str_enc(err, 'utf-8'))
    return ('bytestring', out)



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""CIB helpers, mainly used in the filter definitions"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from .error import ClufterError


PATH_CIB = '/var/lib/pacemaker/cib/cib.xml'


class ResourceSpecError(ClufterError):
    pass


class ResourceSpec(object):
    """Representation of simplified resource specification"""
    # see also pcs/resource.py:resource_list_available
    #                          get_full_ra_type
    # note however that Pacemaker internally prefers this notation:
    # class + class == "ocf"?("::" + provider):"" + ":" + type
    # (i.e., double colon), as per
    # pacemaker/lib/pengine/native.c:{native_print_xml,native_print,
    #                                 get_rscs_brief}
    def __init__(self, spec):
        split = spec.replace('::', ':', 1).split(':', 3)
        try:
            self._type = split.pop()
            maybe_class = split.pop()
            self._class = maybe_class if not(split) else split.pop()
            self._provider = maybe_class if self._class == 'ocf' else None
        except IndexError:
            raise ResourceSpecError("Invalid spec: {0}".format(spec))

    @property
    def res_class(self):
        return self._class

    @property
    def res_provider(self):
        if self._class != 'ocf':
            raise ResourceSpecError("No provider for class `ocf'")
        return self._provider

    @property
    def res_type(self):
        return self._type

    @property
    def xsl_attrs(self):
        ret = ('<xsl:attribute name="class">{0}</xsl:attribute>'
               .format(self._class))
        if self._provider is not None:
            ret += ('<xsl:attribute name="provider">{0}</xsl:attribute>'
                    .format(self._provider))
        ret += ('<xsl:attribute name="type">{0}</xsl:attribute>'
                .format(self._type))
        return ret

    @property
    def xsl_attrs_select(self):
        ret = "@class='{0}'".format(self._class)
        if self._provider is not None:
            ret += " and @provider='{0}'".format(self._provider)
        ret += " and @type='{0}'".format(self._type)
        return ret


def rg2hb_xsl(dst, src=None, req=False, op=False):
    """Emit XSL snippet yielding nvpair-encoded HB RA parameter from RG one

    Parameters:
        dst         destination parameter (name in the nvpair)
        src         source parameter (if not provided, use dst ~ 1:1 match)
        req         valid values: False, True, abs (use raw `src` instead),
                    or None (propage parameter only if non-empty)
                    or Warning (deprecated paramater, implies False
                    and emitted warning [thus the literal object]),
        op          whether this is relevant for "operations" section
    """
    assert req in (False, True, abs, None, Warning), "Invalid `req` param"
    src = src or dst
    deprecated, req = (True, False) if req is Warning else (False, req)
    return (('''\
            <xsl:if test="@{src}">
''' if req is False else '''\
            <xsl:if test="@{src} != ''">
''' if req is None else '') + ('''\
            <xsl:message
            >WARNING: parameter/op {dst} (converted from {src}) is now deprecated</xsl:message>
''' if deprecated else '') + (('''\
            <!-- {dst} ~ {src} -->
            <nvpair id="{{concat($Prefix, '-ATTRS-{dst}')}}"
                    name="{dst}"
''' + ('''\
                    value="{{@{src}}}"/>
''' if req is not abs else '''\
                    value="{src}"/>
''')) if not op else ('''\
            <!-- op:{dst} ~ {src} -->
            <op id="{{concat($Prefix, '-OP-{dst}')}}"
                name="{dst}"
                interval="0"
''' + ('''\
                timeout="{{concat(@{src}, 's')}}"/>
''' if req is not abs else '''\
                timeout="{src}"/>
'''))) + ('''\
            </xsl:if>
''' if not req else '')).format(dst=dst, src=src)



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""ccs2ccs-pcmk filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..filter import XMLFilter


# could be static output, but stay with XSLT in case we want to move sth.
# from cluster.conf to corosync.conf already (used to be a case);
# there is also a possibility to put service-only directive
# as a file under /etc/corosync/service.d/ (called, e.g., pcmk,
# see Clusters from Scratch document)
@XMLFilter.deco_xslt('ccs', 'ccs')
class ccs2ccs_pcmk: pass



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""ccs-disable-rg command"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..command import Command
from ..protocol import protocols
from ..utils_cman import PATH_CLUSTERCONF


@Command.deco(('ccs-disable-rg',
                   ('ccs-version-bump')))
def ccs_disable_rg(cmd_ctxt,
                   input=PATH_CLUSTERCONF,
                   output="cluster-disabledrg-{ccs-disable-rg.in.hash}.conf",
                   noversionbump=False):
    """Make config. prevent RGManager from (accidentally) starting

    Options:
        input          input CMAN-based cluster configuration file
        output         output file with RGManager being declaratively disabled
        noversionbump  prevent incrementing in-document configuration version
    """
    if noversionbump:
        cmd_ctxt['filter_noop'].append('ccs-version-bump')

    file_proto = protocols.plugins['file'].ensure_proto
    return (
        file_proto(input),
        (
            file_proto(output),
        ),
    )



# -*- coding: UTF-8 -*-
# Copyright 2016 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)


ccs2needlexml = '''\
    <node>
        <xsl:for-each select="@*">
            <xsl:variable name="attr_name">
                <xsl:choose>
                    <!-- @nodeid -> @nodeid (unchanged) -->
                    <xsl:when test="name() = 'nodeid'">
                        <xsl:value-of select="name()"/>
                    </xsl:when>
                    <!-- @name -> @ring0_addr -->
                    <xsl:when test="name() = 'name'">
                        <xsl:value-of select="'ring0_addr'"/>
                    </xsl:when>
                    <!-- @votes -> @quorum_votes -->
                    <xsl:when test="name() = 'votes'">
                        <!-- xsl:value-of select="'quorum_votes'"/ -->
                        <xsl:value-of select="''"/>
                        <xsl:message>
                            <xsl:value-of select="concat('WARNING: node ',
                                                        ../@nodeid,
                                                        ' declared with ',
                                                        .,
                                                        ' vote(s), which is ')"/>
                                <xsl:choose>
                                    <xsl:when test="number(.) = 1">
                                        <xsl:value-of select="'a default anyway,'"/>
                                    </xsl:when>
                                    <xsl:otherwise>
                                        <xsl:value-of select="concat('something not',
                                                                     ' 100% sane',
                                                                     ' nor advisable,')"/>
                                    </xsl:otherwise>
                                </xsl:choose>
                            <xsl:value-of select="concat(' hence not propagated (as',
                                                         ' `quorum_votes` property)')"/>
                        </xsl:message>
                    </xsl:when>
                    <xsl:otherwise>
                        <xsl:value-of select="''"/>
                    </xsl:otherwise>
                </xsl:choose>
            </xsl:variable>
            <xsl:if test="$attr_name != ''">
                <xsl:attribute name="{$attr_name}">
                    <xsl:value-of select="."/>
                </xsl:attribute>
            </xsl:if>
        </xsl:for-each>
        <xsl:if test="altname/@name">
            <xsl:attribute name="ring1_addr">
                <xsl:value-of select="altname/@name"/>
            </xsl:attribute>
        </xsl:if>
    </node>
'''

ccs2ccs_pcmk = '''\
    <xsl:copy>
        <xsl:copy-of select="@*|altname"/>
        <!-- "unfence" section disappears, "fence" one enforced below -->
        <fence>
            <method name="pcmk-method">
                <device name="pcmk-redirect" port="{@name}"/>
            </method>
        </fence>
    </xsl:copy>
'''

# following could be omitted but keep it around if we ever need
# to add some node attributes in the future
ccsflat2cibprelude = '''\
    <!-- differentiate between RHEL 6 and higher (rhbz#1207345) -->
    <node>
        <xsl:attribute name="id">
            <xsl:choose>
                <xsl:when test="$pcscmd_cman">
                    <!-- pacemaker + cman -->
                    <xsl:value-of select="@name"/>
                </xsl:when>
                <xsl:otherwise>
                    <!-- pacemaker + corosync/needle -->
                    <xsl:value-of select="@nodeid"/>
                </xsl:otherwise>
            </xsl:choose>
        </xsl:attribute>
        <xsl:attribute name="uname">
            <xsl:value-of select="@name"/>
        </xsl:attribute>
        <xsl:attribute name="type">member</xsl:attribute>
    </node>
'''

###

ccspcmk2pcscmd = '''\
    <xsl:value-of select="concat(' ', @name)"/>
    <xsl:if test="altname/@name">
        <xsl:value-of select="concat(',', altname/@name)"/>
    </xsl:if>
'''



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""simpleconfig2needlexml filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from lxml import etree

from ..filter import Filter
from ..utils import selfaware


@selfaware
def _simpleconfig2needlexml(me, element, options, sections):
    # must not attempt to modify anything from options/sections in-place
    element.attrib.update(options)
    element.extend([me(etree.Element(s[0]), *s[1:]) for s in sections])
    return element


@Filter.deco('simpleconfig-normalized', 'coroxml-needle')
def simpleconfig2needlexml(flt_ctxt, in_obj):
    """Cooks XML representation of corosync.conf"""
    struct = in_obj('struct', protect_safe=True)
    return ('etree', _simpleconfig2needlexml(etree.Element("corosync"),
                                             *struct[1:]))



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

from ....utils_xslt import xslt_is_member

###

ccs2needlexml_attrs = (
    'ringnumber',
    'bindnetaddr',
    'broadcast',
    'mcastaddr',
    'mcastport',
    'ttl',
)

ccs2needlexml = '''\
    <xsl:copy>
        <xsl:copy-of select="@*[
''' + (
            xslt_is_member('name()', ccs2needlexml_attrs)
) + ''']" />
    </xsl:copy>
'''

###

ccspcmk2pcscmd = '''\
    <xsl:variable name="RingNumber">
        <xsl:choose>
            <xsl:when test="@ringnumber = 0
                            or
                            (
                              not(@ringnumber)
                              and
                              count(preceding-sibling::interface) = 0
                            )">
                <xsl:value-of select="'0'"/>
            </xsl:when>
            <xsl:when test="@ringnumber = 1
                            or
                            (
                              not(@ringnumber)
                              and
                              count(preceding-sibling::interface) = 1
                            )">
                <xsl:value-of select="'1'"/>
            </xsl:when>
        </xsl:choose>
    </xsl:variable>
    <xsl:if test="$RingNumber">
        <xsl:if test="@bindnetaddr">
            <xsl:value-of select="concat(' --addr', $RingNumber,' ', @bindnetaddr)"/>
        </xsl:if>
        <xsl:if test="@mcastaddr">
            <xsl:value-of select="concat(' --mcast', $RingNumber,' ', @mcastaddr)"/>
        </xsl:if>
        <xsl:if test="@mcastport">
            <xsl:value-of select="concat(' --mcastport', $RingNumber,' ', @mcastport)"/>
        </xsl:if>
        <xsl:if test="@broadcast">
            <xsl:value-of select="concat(' --broadcast', $RingNumber)"/>
        </xsl:if>
    </xsl:if>
'''



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

# following 2nd chance import is to allow direct usage context (testing, etc.)
try:
    from ....utils_xslt import xslt_is_member
except ValueError:  # Value?
    from ...utils_xslt import xslt_is_member

###

# XXX: normally, there would be fs.py containing this common implementation
#      and {cluster,net}.py referring to it, but due to the way it is used
#      in the parent, current approach seems better

try:
    from ....utils_cib import ResourceSpec
except ValueError:
    from ...utils_cib import ResourceSpec

ccsflat2cibprelude = ('''
    <!--
        Filesystem ~ {,cluster,net}fs
     -->
    <xsl:when test="name() = '%(target)s'">
        <xsl:variable name="FsKind" select="name()"/>
''' + (
        ResourceSpec('ocf:heartbeat:Filesystem').xsl_attrs
) + '''
        <!-- INSTANCE_ATTRIBUTES -->
        <instance_attributes id="{concat($Prefix, '-ATTRS')}">
            <!-- device ~ device -->
            <nvpair id="{concat($Prefix, '-ATTRS-device')}"
                    name="device">
                <xsl:attribute name="value">
                    <xsl:choose>
                        <!-- LABEL or UUID specification -->
                        <xsl:when test="starts-with(@device, 'LABEL=')">
                            <xsl:value-of select="concat(
                                                      '-L',
                                                      substring-after(@device, 'LABEL=')
                                                  )"/>
                        </xsl:when>
                        <xsl:when test="starts-with(@device, 'UUID=')">
                            <xsl:value-of select="concat(
                                                      '-U',
                                                      substring-after(@device, 'UUID=')
                                                  )"/>
                        </xsl:when>
                        <!-- specification a la netfs-->
                        <xsl:when test="$FsKind = 'netfs'
                                        and
                                        (
                                            not(@fstype)
                                            or
                                            starts-with(@fstype, 'nfs')
                                        )">
                            <xsl:value-of select="concat(
                                                      @host,
                                                      ':',
                                                      @export)"/>
                        </xsl:when>
                        <xsl:when test="$FsKind = 'netfs'
                                        and
                                        starts-with(@fstype, 'cifs')">
                            <xsl:value-of select="concat(
                                                      '//',
                                                      @host,
                                                      '/',
                                                      @export)"/>
                        </xsl:when>
                        <xsl:when test="$FsKind = 'bind-mount'">
                            <xsl:value-of select="@source"/>
                        </xsl:when>
                        <xsl:otherwise>
                            <xsl:value-of select="@device"/>
                        </xsl:otherwise>
                    </xsl:choose>
                </xsl:attribute>
            </nvpair>
            <!-- directory ~ mountpoint -->
            <nvpair id="{concat($Prefix, '-ATTRS-directory')}"
                    name="directory"
                    value="{@mountpoint}"/>
            <!-- fstype ~ fstype -->
            <xsl:if test="@fstype or $FsKind = 'bind-mount'">
            <nvpair id="{concat($Prefix, '-ATTRS-fstype')}"
                    name="fstype">
                <xsl:attribute name="value">
                    <xsl:choose>
                        <xsl:when test="$FsKind = 'bind-mount'">
                            <xsl:value-of select="'none'"/>
                        </xsl:when>
                        <xsl:otherwise>
                            <xsl:value-of select="@fstype"/>
                        </xsl:otherwise>
                    </xsl:choose>
                </xsl:attribute>
            </nvpair>
            </xsl:if>
            <!-- options ~ options -->
            <xsl:if test="@options
                          or
''' + (
                          xslt_is_member('$FsKind', ('netfs', 'bind-mount'))
) + '''">
            <nvpair id="{concat($Prefix, '-ATTRS-options')}"
                    name="options">
                <xsl:attribute name="value">
                    <xsl:choose>
                        <xsl:when test="$FsKind = 'bind-mount'">
                            <xsl:value-of select="'bind'"/>
                        </xsl:when>
                        <xsl:when test="@options">
                            <xsl:value-of select="@options"/>
                        </xsl:when>
                        <xsl:when test="$FsKind = 'netfs'
                                        and
                                        (
                                            not(@fstype)
                                            or
                                            starts-with(@fstype, 'nfs')
                                        )">
                            <xsl:value-of select="'sync,soft,noac'"/>
                        </xsl:when>
                        <xsl:when test="$FsKind = 'netfs'
                                        and
                                        starts-with(@fstype, 'cifs')">
                            <xsl:value-of select="'guest'"/>
                        </xsl:when>
                        <xsl:otherwise>
                            <xsl:message terminate="yes"
                            >Unexpected *fs RA params combination witnessed.</xsl:message>
                        </xsl:otherwise>
                    </xsl:choose>
                </xsl:attribute>
            </nvpair>
            </xsl:if>
            <!-- run_fsck ~ force_fsck -->
            <xsl:if test="$FsKind != 'bind-mount'
                          and
''' + (
                xslt_is_member('@force_fsck', ('1', 'yes'))
) + '''">
            <nvpair id="{concat($Prefix, '-ATTRS-run_fsck')}"
                    name="run_fsck"
                    value="force"/>
            </xsl:if>
            <!-- force_unmount (true|safe|false) ~ force_unmount (true...|false...)
                 - this depends on
                   https://github.com/ClusterLabs/resource-agents/pull/423
                 - alternatively the commented out part below might be used
                   to achieve similar behavior (no signal ever sent) -->
            <nvpair id="{concat($Prefix, '-ATTRS-force_unmount')}"
                    name="force_unmount"
                    value="false">
                <xsl:if test="not(
''' + (
                    xslt_is_member('@force_unmount', ('1', 'yes', 'on', 'true'))
) + ''')">
                    <xsl:attribute name="value">
                        <xsl:value-of select="'safe'"/>
                    </xsl:attribute>
                </xsl:if>
            </nvpair>
            </instance_attributes>
        <!-- xsl:if test="not(contains(concat(
                          '|yes',
                          '|on',
                          '|true',
                          '|1',
                          '|'), concat('|', @force_unmount, '|')))">
        < operations>
            - this should cause no signal is ever sent to FS users -
            <op id="{concat($Prefix, '-OPS-shutdown')}"
                name="stop"
                interval="0"
                timeout="1"/>
        </operations>
        </xsl:if -->
    </xsl:when>
''') % dict(target=__name__.rsplit('.', 1)[-1].split('_', 1)[-1])



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....filters.ccs_artefacts import artefact_cond_ra

ccs_artefacts = ''.join((
    artefact_cond_ra('@config_file',
                     kind='A', desc='configuration file'),
))

###

from ....utils_cib import ResourceSpec, rg2hb_xsl

ccsflat2cibprelude = '''\
    <!--
        pgsql ~ postgres-8
     -->
    <xsl:when test="name() = 'postgres-8'">
''' + (
        ResourceSpec('ocf:heartbeat:pgsql').xsl_attrs
) + '''
        <!-- INSTANCE_ATTRIBUTES -->
        <instance_attributes id="{concat($Prefix, '-ATTRS')}">
''' + (
            rg2hb_xsl('config', 'config_file', req=True)
            +
            rg2hb_xsl('start_opt', 'postmaster_options')
            +
            rg2hb_xsl('pgdba', 'postmaster_user')
) + '''\
            <!-- XXX some items from postmaster_options could be
                 re-parsed into respective RA params
                 (-h $OCF_RESKEY_pghost) -->
        </instance_attributes>

        <!-- OPERATIONS -->
        <operations>
''' + (
            rg2hb_xsl('start', 'startup_wait', op=True)
) + '''\
        </operations>
    </xsl:when>
'''



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....utils_cib import ResourceSpec, rg2hb_xsl


ccsflat2cibprelude = '''\
    <!--
        VirtualDomain ~ vm
     -->
    <xsl:when test="name() = 'vm'">
''' + (
        ResourceSpec('ocf:heartbeat:VirtualDomain').xsl_attrs
) + '''
        <!-- SHOW-STOPPERS -->
        <xsl:if test="@use_virsh = '0'">
            <xsl:message terminate="yes"
            >Cannot reliably convert non-virsh configuration</xsl:message>
        </xsl:if>

        <!-- INSTANCE_ATTRIBUTES -->
        <instance_attributes id="{concat($Prefix, '-ATTRS')}">
            <!-- config ~ xmlfile -->
            <xsl:if test="@xmlfile or @path">
            <nvpair id="{concat($Prefix, '-ATTRS-config')}"
                    name="config"
                    value="{@xmlfile}">
                <xsl:if test="not(@xmlfile)">
                    <xsl:attribute name="value">
                        <!-- XXX simply assume the extension -->
                        <xsl:value-of select="concat(@path, '/', @name, '.xml')"/>
                    </xsl:attribute>
                </xsl:if>
            </nvpair>
            </xsl:if>
''' + (
            rg2hb_xsl('hypervisor', 'hypervisor_uri')
            +
            rg2hb_xsl('snapshot', req=None)
) + '''\
        </instance_attributes>

        <!-- OPERATIONS -->
        <xsl:variable name="VmStop"
                      select="action[
                                  @name = 'stop'
                                  and
                                  number(@timeout) &gt; 0
                              ]"/>
        <xsl:if test="$VmStop">
            <operations>
''' + (
                rg2hb_xsl("stop", "{concat($VmStop/@timeout, 's')}",
                          req=abs, op=True)
) + '''\
            </operations>
        </xsl:if>

    </xsl:when>
'''

###

from ....filters.ccs_artefacts import artefact_cond_ra

ccs_artefacts = ''.join((
    artefact_cond_ra('@xmlfile',
                     kind='A', desc='libvirt XML domain definition'),
))



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Various little+independent helpers"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from itertools import takewhile
from types import GeneratorType

from .utils_2to3 import PY3, basestring, iter_items


# inspired by http://stackoverflow.com/a/4374075
if PY3:
    immutable = lambda x: isinstance(x, (basestring, int, bool, float, tuple,
                                         GeneratorType))
else:
    immutable = lambda x: isinstance(x, (basestring, int, long, bool, float,
                                         tuple, GeneratorType))

tuplist = lambda x: isinstance(x, (tuple, list, set))
# turn args into tuple unless single tuplist arg
args2sgpl = \
    lambda x=(), *y: x if not y and tuplist(x) else (x, ) + y
arg2wrapped = \
    lambda x=(), *y: \
        x if not y and isinstance(x, tuple) else (x, ) + y
args2unwrapped = \
    lambda x=None, *y: x if not y else (x, ) + y
# turn args into tuple unconditionally
args2tuple = lambda *args: args
any2iter = \
    lambda x: \
        x if hasattr(getattr(x, 'next', getattr(x, '__next__', None)),
                     '__call__') \
        else iter(args2sgpl(x, None))

head_tail = \
    lambda x=None, *y, **kwargs: \
        (x, args2sgpl(*y)) if not tuplist(x) or kwargs.get('stop', 0) \
                           else (head_tail(stop=1, *tuple(x) + y))

nonetype = type(None)

identity = lambda x: x


#
# easily mangle with dicts/sets (eventually lists), returning dicts
#

filterdict_map = \
    lambda fn, src, *which, **update: \
        dict(((x, fn(x)) for x in which if x in src), **update)

# .copy() so as to allow for in-situ manipulations like .pop() without
# affecting the running iteration
filterdict_invmap = \
    lambda fn, src, *which, **update: \
        dict(((x, fn(x)) for x in src.copy() if x not in which), **update)

#

filterdict_keep = \
    lambda src, *which, **update: \
        filterdict_map(
            lambda x, fn=update.pop('_fn_', identity): fn(src[x]),
            src, *which, **update
        )

filterdict_invkeep = \
    lambda src, *which, **update: \
        filterdict_invmap(
            lambda x, fn=update.pop('_fn_', identity): fn(src[x]),
            src, *which, **update
        )

#

filterdict_pop = \
    lambda src, *which, **update: \
        filterdict_map(
            lambda x, fn=update.pop('_fn_', identity): fn(src.pop(x)),
            src, *which, **update
        )

filterdict_invpop = \
    lambda src, *which, **update: \
        filterdict_invmap(
            lambda x, fn=update.pop('_fn_', identity): fn(src.pop(x)),
            src, *which, **update
        )

# following supposes that remove() returns None (as is the case with set/list)

filterdict_remove = \
    lambda src, *which, **update: \
        filterdict_map(
            lambda x, fn=update.pop('_fn_', identity): fn(src.remove(x) or x),
            src, *which, **update
        )

filterdict_invremove = \
    lambda src, *which, **update: \
        filterdict_invmap(
            lambda x, fn=update.pop('_fn_', identity): fn(src.remove(x) or x),
            src, *which, **update
        )


#
# introspection related
#

def isinstanceexcept(subj, obj, *exc):
    return isinstance(subj, obj) and not isinstance(subj, exc)


def isinstanceupto(subj, obj, *exc):
    return isinstance(subj,
                      tuple(takewhile(lambda x: x not in exc, obj.__mro__)))


def areinstances(obj1, obj2):
    return isinstance(obj1, obj2.__class__) or isinstance(obj2, obj1.__class__)


def areinstancesupto(obj1, obj2, *exc):
    return isinstanceupto(obj1, obj2.__class__, *exc) \
        or isinstanceupto(obj2, obj1.__class__, *exc)


def popattr(obj, what, *args):
    assert len(args) < 2
    ret = getattr(obj, what, *args)
    try:
        delattr(obj, what)
    except AttributeError:
        if args:
            ret = args[0]
        else:
            raise
    return ret


def iterattrs(obj, skip_private=True):
    """Iterate through (unbound) attributes of obj, skipping private or not"""
    if skip_private:
        return ((n, v) for n, v in iter_items(obj.__dict__)
                if not n.startswith('__'))
    return iter_items(obj.__dict__)


def func_defaults_varnames(func, skip=0):
    """Using introspection, get arg defaults (dict) + all arg names (tuple)

    Parameters:
        skip                how many initial arguments to skip
    """
    code = func.__code__
    func_varnames = code.co_varnames[skip:code.co_argcount]

    func_defaults = dict(zip(
        reversed(func_varnames),
        reversed(func.__defaults__),
    ))

    return func_defaults, func_varnames


#
# decorators
#

def selfaware(func):
    """Decorator suitable for recursive staticmethod"""
    def selfaware_inner(*args, **kwargs):
        return func(selfaware(func), *args, **kwargs)
    for attr in ('__doc__', '__name__'):
        setattr(selfaware_inner, attr, getattr(func, attr))
    return selfaware_inner


# Inspired by http://stackoverflow.com/a/1383402
class classproperty(property):
    def __init__(self, fnc):
        property.__init__(self, classmethod(fnc))

    def __get__(self, this, owner):
        return self.fget.__get__(None, owner)()


class hybridproperty(property):
    def __init__(self, fnc):
        property.__init__(self, classmethod(fnc))

    def __get__(self, this, owner):
        return self.fget.__get__(None, this if this else owner)()


class hybridmethod(property):
    def __init__(self, fnc):
        property.__init__(self, classmethod(fnc))

    def __get__(self, this, owner):
        return self.fget.__get__(None, this if this else owner)


# inspired from speaklater: http://pypi.python.org/pypi/speaklater
class lazystring(object):
    """Mimic string that in fact is on-off constructed on-demand

    Note: only direct use, str methods and '+' operator supported.
    """
    def __init__(self, fnc, cache=True):
        self._fnc = fnc
        self._cache = cache

    @property
    def content(self):
        ret = self._cache
        if ret is False or ret is True:
            ret = self._fnc()
            if ret is True:
                self._cache = ret
        return ret

    def __str__(self):           return str(self.content)
    def __repr__(self):          return repr(self.content)
    def __getattr__(self, what): return getattr(self.content, what)
    def __add__(self, other):    return self.content + other
    def __radd__(self, other):   return other + self.content



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""CMAN helpers, mainly used in the filter definitions"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

PATH_CLUSTERCONF = '/etc/cluster/cluster.conf'


def get_nodes(etree):
    return etree.xpath('/cluster/clusternodes/clusternode')



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

###

from ....utils_xslt import xslt_is_member

# 1:1 mapping of supported params to pcs arguments
needlexml2pcscmd_supported = (
    # binary 0|1 values
    'auto_tie_breaker',
    'last_man_standing',
    'wait_for_all',
    # standard values
    'last_man_standing_window',
)

needlexml2pcscmd = '''\
    <xsl:for-each select="@*[
''' + (
        xslt_is_member('name()', needlexml2pcscmd_supported)
) + ''']">
        <xsl:if test=".">
            <!-- but see https://bugzilla.redhat.com/1235452 -->
            <xsl:value-of select="concat(' --', name(), '=', .)"/>
        </xsl:if>
    </xsl:for-each>
'''



# -*- coding: UTF-8 -*-
# Copyright 2018 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

###

from ....filters._2pcscmd import verbose_ec_test, verbose_inform
from ....utils_xslt import NL, xslt_is_member


needlexml2pcscmd_attrs = (
    'algorithm',  # required
    'connect_timeout',
    'force_ip_version',
    'host',  # required
    'port',
    'tie_breaker',
    'tls',  # see https://bugzilla.redhat.com/1476862
)

needleqdevicexml2pcscmd = '''\
    <xsl:when test="@model = 'net' and $pcscmd_extra_qnet">
        <xsl:for-each select="net/@*[
''' + (
            xslt_is_member('name()', needlexml2pcscmd_attrs)
) + ''']">
            <xsl:choose>
                <xsl:when test="name() = 'tls'
                                and
                                . = 'on'">
                </xsl:when>
                <xsl:when test="name() = 'tls'">
                    <xsl:message>
                        <xsl:value-of select="concat('WARNING: non-default',
                                                     ' `quorum.device.net.tls`',
                                                     ' value `', ., '` specified,',
                                                     ' but current pcs not capable',
                                                     ' to set it')"/>
                    </xsl:message>
                </xsl:when>
                <xsl:otherwise>
                    <xsl:value-of select="concat(' ', name(), '=', .)"/>
                </xsl:otherwise>
            </xsl:choose>
        </xsl:for-each>
    </xsl:when>
    <xsl:when test="@model = 'net'">
        <xsl:value-of select="'clufter:UNSUPPORTED'"/>
    </xsl:when>
'''



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

from __future__ import print_function

"""Command manager"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from logging import getLogger
from os.path import abspath, dirname, join
from textwrap import wrap
from sys import modules

from .command import commands, CommandAlias
from .error import ClufterError, ClufterPlainError, \
                   EC
from .filter_manager import FilterManager
from .plugin_registry import PluginManager
from .utils import filterdict_keep
from .utils_2to3 import basestring, iter_items, iter_values
from .utils_func import apply_intercalate, bifilter_unpack
from .utils_prog import make_options, set_logging

log = getLogger(__name__)


class CommandManagerError(ClufterError):
    pass


class CommandNotFoundError(ClufterPlainError):
    def __init__(self, cmd):
        super(CommandNotFoundError, self).__init__("Command not found: `{0}'",
                                                   cmd)


class CommandManager(PluginManager):
    """Class responsible for commands routing to filters or other actions"""
    _default_registry = commands

    @classmethod
    def _init_plugins(cls, commands, flt_mgr=None, *args, **kwargs):
        log.debug("Commands before resolving: {0}".format(commands))
        if flt_mgr is None:
            flts = set()
            for cmd in iter_values(commands):
                for flt in apply_intercalate(getattr(cmd, 'filter_chain', ())):
                    flts.add(flt)
            flt_mgr = FilterManager.init_lookup(flts, **kwargs)
        return cls._resolve(flt_mgr.filters, commands, *args,
                            **filterdict_keep(kwargs, 'system', 'system_extra'))

    @staticmethod
    def _resolve(filters, commands, system='', system_extra=''):
        # name -> (cmd obj if not alias or resolvable name)
        aliases = []
        inverse_commands = dict((b, a) for a, b in iter_items(commands))

        # first, resolve end-use commands
        for cmd_name, cmd_cls in commands.items():
            if issubclass(cmd_cls, CommandAlias):
                aliases.append(cmd_name)
                continue
            ret = cmd_cls(filters)
            if ret is not None:
                commands[cmd_name] = ret
            else:
                commands.pop(cmd_name)

        # only then resolve the command aliases, track a string identifying
        # end-use command in `commands`
        for cmd_name in aliases:
            alias = commands[cmd_name]
            if not isinstance(alias, basestring):
                # not resolved yet
                assert issubclass(alias, CommandAlias)
                resolved = alias(filters, commands, inverse_commands,
                                 system, system_extra)
                resolved_cls = type(resolved)
                if resolved_cls not in inverse_commands:
                    if resolved is not None:
                        log.warning("Resolve at `{0}' alias: target unknown"
                                    .format(cmd_name))
                    commands.pop(cmd_name)
                else:
                    commands[cmd_name] = inverse_commands[resolved_cls]

        return commands

    @property
    def commands(self):
        return self._plugins

    def __call__(self, parser, args=None):
        """Follow up of the entry point, facade to particular commands"""
        ec = EC.EXIT_SUCCESS
        values = parser.values
        try:
            canonical_cmd = command = cmd = getattr(values, 'help', None) \
                                            or args[0]
            while isinstance(command, basestring):
                canonical_cmd = command
                command = self._plugins.get(command, None)
            if not command:
                # XXX maybe it just hasn't been resolved successfully
                raise CommandNotFoundError(cmd)

            opt_group = parser.option_groups[0]
            parser.description, options = command.parser_desc_opts(opt_group)
            opt_group.add_options(make_options(options))

            args = ['--help'] if values.help else args[1:]
            parser.defaults.update(values.__dict__)  # from global options
            opts, args = parser.parse_args(args)
            if opts.help:
                sep = '\n{0:^{width}}'.format('or:', width=len('Usage: '))
                usage = sep.join(map(
                    lambda c:
                        "{0}{1} [<cmd option ...>]"
                        .format("%prog [<global option> ...] "
                                if parser.prog != c else '', c),
                    sorted(set([cmd, canonical_cmd]),
                           key=lambda i: int(i == canonical_cmd))
                ))
                print(parser.format_customized_help(usage=usage))
                return ec

            set_logging(opts)
            log.debug("Running command `{0}';  opts={1}, args={2}"
                      .format(cmd, opts.__dict__, args))
            ec = command(opts, args)
        except ClufterError as e:
            ec = EC.EXIT_FAILURE
            print(e)
            if isinstance(e, CommandNotFoundError):
                print("\n" + self.pretty_cmds())
        #except Exception as e:
        #    print("OOPS: underlying unexpected exception:\n{0}".format(e))
        #    ec = EC.EXIT_FAILURE
        return ec

    def pretty_cmds(self, text_width=77, linesep_width=1,
                    ind='  ', itemsep='\n', secsep='\n', intmark=' *',
                    cmds_intro='Commands', aliases_intro='Aliases:',
                    ellip_yes='; only built-in ones included:',
                    ellip_no='; those without star (\'*\') are built-in:',
                    ref_str='alias for {0}', ellip=False):
        """Return string containing formatted list of commands (name + desc)"""
        pth = join(dirname(abspath(__file__)), self._registry.__name__)  # int.
        cmds_aliases = [
            ([(name, ref_str.format(obj) if i else obj.__doc__.splitlines()[0],
               i or dirname(modules[obj.__class__.__module__].__file__) == pth)
              for name, obj in sorted(cat) if obj],
              max((0, ) + tuple(len(name) for name, obj in cat
                                if obj and (not ellip or i or dirname(
                                    modules[obj.__class__.__module__].__file__
                                ) == pth)))
            )
            for i, cat in enumerate(bifilter_unpack(
                lambda name, obj: not isinstance(obj, basestring),
                iter_items(self._plugins)
            ))
        ]
        width = max(i[1] for i in cmds_aliases) + linesep_width
        desc_indent = ind + (width * ' ')
        text_width = max(text_width - len(desc_indent), 20)
        cmds_intro += ellip_yes if ellip else ellip_no
        return secsep.join(
            itemsep.join([header] + [
                '{0}{1:{width}}{2}'.format(
                    ind if internal else intmark + ind[len(intmark):],
                    name, '\n'.join(
                        wrap(desc,
                             width=text_width, subsequent_indent=desc_indent)
                    ), width=width
                ) for name, desc, internal in i[0] if not ellip or internal
            ]) for header, i in zip((cmds_intro, aliases_intro), cmds_aliases)
        )



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....filters._2pcscmd import verbose_ec_test, verbose_inform
from ....utils_xslt import NL

cib2pcscmd = ('''\
''' + (
    verbose_inform('"new stonith level: ", @index')
) + '''
    <xsl:value-of select='concat($pcscmd_pcs, "stonith level add",
                                 " ", @index,
                                 " ", @target,
                                 " &apos;", @devices, "&apos;")'/>
    <xsl:value-of select="'%(NL)s'"/>
''' + (
    verbose_ec_test
) + '''
''') % dict(
    NL=NL,
)



# -*- coding: UTF-8 -*-
# Copyright 2018 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Utility functions wrt. cluster systems in general"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from functools import reduce
try:
    from itertools import chain, zip_longest
except ImportError:  # PY2 backward compatibility
    from itertools import chain, izip_longest as zip_longest
from logging import getLogger
from operator import add

from .error import ClufterPlainError
from .utils import args2sgpl
from .utils_2to3 import basestring, enumerate_u, iter_items, iter_values, \
                        reduce_u
from .utils_func import apply_intercalate

log = getLogger(__name__)

class FactsError(ClufterPlainError):
    pass

#
# EXECUTABLE KNOWLEDGE ABOUT THE CLUSTER PACKAGES PER SYSTEM/DISTROS
#
# or a bit of a logic programming paradigm with Python...
#
# ... first, define some facts (in suitable data structures):
#

# core hierarchical (sparse!) map:
# system -> distro -> release -> package -> version
#
# The notation should be self-explanatory, perhaps except 'component[extra]'
# (borrowed from the very similar construct of setuptools, there can be more
# "extra traits" delimited with comma) ... simply 'pacemaker[cman]' means that
# such pacemaker is somehow associated with "cman" (more specificially, it is
# intended [compilation flags, etc.] to be used with cman), which apparently
# doesn't match 'pacemaker[coro]' component specification in the input query.
cluster_map = {
    'linux':
        {
            'debian': (
                ((6, ), {
                    # https://packages.debian.org/squeeze/$PACKAGE
                    'corosync':                      (1, 2),
                    'pacemaker[+coro,+hb]':          (1, 0, 9),
                    #---
                    'sys::init-sys':                'sysvinit',
                }),
                ((7, ), {
                    # https://packages.debian.org/wheezy/$PACKAGE
                    'corosync':                      (1, 4),
                    'pacemaker[+coro,+hb]':          (1, 1, 7),
                    'resource-agents':               (3, 9, 2),
                }),
                ((8, ), {
                    # https://packages.debian.org/jessie/$PACKAGE
                    'corosync':                      (1, 4, 6),
                    'resource-agents':               (3, 9, 3),
                    #---
                    'sys::init-sys':                'systemd',
                }),
                ((9, ), {
                    # https://packages.debian.org/stretch/$PACKAGE
                    'corosync':                      (2, 4, 2),
                    'pacemaker[+coro]':              (1, 1, 16),
                    'pcs':                           (0, 9, 155),
                    'resource-agents':               (4, 0, 0),  # rc1
                }),
                # currently a moving target
                ((10, ), {
                    # https://packages.debian.org/buster/$PACKAGE
                    'pacemaker[+coro]':              (1, 1, 18),  # rc3
                    'pcs':                           (0, 9, 161),
                    'resource-agents':               (4, 0, 1),
                }),
            ),
            'fedora': (
                # see also
                # https://fedoraproject.org/wiki/Releases/$RELEASE/Schedule
                ((13, ), {
                    'corosync':                      (1, 3),
                    'pacemaker[+cman]':              (1, 1, 4),
                    #---
                    'pkg::mysql':                   'mysql-server',
                    #---
                    'cmd::pkg-install':             'yum install -y {packages}',
                    #---
                    'sys::init-sys':                'upstart',
                }),
                ((14, ), {
                    'corosync':                      (1, 4),
                    'pacemaker[+cman]':              (1, 1, 6),
                }),
                ((15, ), {
                    'sys::init-sys':                'systemd',
                }),
                ((17, ), {
                    'corosync':                      (2, 3),
                    'pacemaker[+coro]':              (1, 1, 7),
                    'pcs':                           (0, 9, 1),
                    #'pcs':                           (0, 9, 3),  # updates
                }),
                ((18, ), {
                    'pacemaker[+coro]':              (1, 1, 8),
                    'pcs':                           (0, 9, 27),
                }),
                ((19, ), {
                    'pacemaker[+coro]':              (1, 1, 9),
                    'pcs':                           (0, 9, 36),
                    #'pcs':                           (0, 9, 48),  # updates
                    'resource-agents':               (3, 9, 5),
                    #---
                    # https://fedoraproject.org/wiki/Features/ReplaceMySQLwithMariaDB
                    'pkg::mysql':                   'mariadb-server',
                }),
                ((20, ), {
                    'pcs':                           (0, 9, 44),
                    #'pcs':                           (0, 9, 102),  # updates
                    #'pcs':                           (0, 9, 115),  # updates
                }),
                ((21, ), {
                    'pacemaker[+coro]':              (1, 1, 12),
                    #'pacemaker[+coro]':              (1, 1, 13),  # updates
                    'pcs':                           (0, 9, 115),
                    #'pcs':                           (0, 9, 137),  # updates
                }),
                ((22, ), {
                    #'pacemaker[+coro]':              (1, 1, 13),  # updates
                    'pcs':                           (0, 9, 139),
                    #'pcs':                           (0, 9, 149),  # updates
                    'resource-agents':               (3, 9, 6),
                    #---
                    # https://fedoraproject.org/wiki/Changes/ReplaceYumWithDNF
                    'cmd::pkg-install':             'dnf install -y {packages}',
                }),
                ((23, ), {
                    'pacemaker[+coro]':              (1, 1, 13),
                    #'pacemaker[+coro]':              (1, 1, 14),  # updates
                    #'pacemaker[+coro]':              (1, 1, 15),  # updates
                    'pcs':                           (0, 9, 144),
                    #'pcs':                           (0, 9, 149),  # updates
                }),
                ((24, ), {
                    #'corosync':                      (2, 4),  # updates
                    'pacemaker[+coro]':              (1, 1, 14),
                    #'pacemaker[+coro]':              (1, 1, 15),  # updates
                    'pcs':                           (0, 9, 150),
                    'resource-agents':               (3, 9, 7),
                }),
                ((25, ), {
                    'corosync':                      (2, 4, 1),
                    #'pacemaker[+coro]':              (1, 1, 16),  # updates
                    'pcs':                           (0, 9, 154),
                }),
                ((26, ), {
                    'corosync':                      (2, 4, 2),
                    'pacemaker[+coro]':              (1, 1, 17),
                    'pcs':                           (0, 9, 156),
                    'resource-agents':               (4, 0, 1),
                }),
                ((27, ), {
                    #'corosync':                      (2, 4, 3),  # updates
                    #'pacemaker[+coro]':              (1, 1, 18), # updates
                    'pcs':                           (0, 9, 159),
                }),
                ((28, ), {
                    'corosync':                      (2, 4, 3),
                    'pacemaker[+coro]':              (1, 1, 18),
                    'pcs':                           (0, 9, 160),
                }),
                # rawhide, i.e., moving target...
                #((28, 999), {
                #}),
                # coming...
                #((29, ), {
                #}),
            ),
            'redhat': (
                ((6, 0), {
                    'corosync':                      (1, 2),
                    #---
                    'pkg::lvm':                     'lvm2',
                    'pkg::mysql':                   'mysql-server',
                    'pkg::postgresql':              'postgresql-server',
                    'pkg::tomcat':                  'tomcat6',
                    'pkg::virsh':                   'libvirt-client',
                    #---
                    'cmd::pkg-install':             'yum install -y {packages}',
                    #---
                    'sys::init-sys':                'upstart',
                }),
                ((6, 2), {
                    'corosync':                      (1, 4),
                    # first time the upstream switch to HB agents happened
                    'resource-agents':               (3, 9, 2),
                }),
                ((6, 4), {
                    'pcs':                           (0, 9, 26),
                }),
                ((6, 5), {
                    'pacemaker[+cman]':              (1, 1, 10),
                    'pcs':                           (0, 9, 90),
                }),
                ((6, 6), {
                    'pacemaker[+cman]':              (1, 1, 12),
                    'pcs':                           (0, 9, 123),
                    'resource-agents':               (3, 9, 5),
                }),
                ((6, 7), {
                    'pacemaker[+cman]':              (1, 1, 12),
                    'pcs':                           (0, 9, 139),
                }),
                ((6, 8), {
                    'pacemaker[+cman]':              (1, 1, 14),
                    'pcs':                           (0, 9, 148),
                }),
                ((6, 9), {
                    'pacemaker[+cman]':              (1, 1, 15),
                    'pcs[-agents-via-pacemaker]':    (0, 9, 155),
                }),
                ((6, 10), {
                    'pacemaker[+cman]':              (1, 1, 18),  # guess
                }),
                ((7, 0), {
                    'corosync':                      (2, 3),
                    'pacemaker[+coro]':              (1, 1, 10),
                    'pcs':                           (0, 9, 115),
                    # may differ from latest 6.x
                    'resource-agents':               (3, 9, 5),
                    #---
                    'pkg::mysql':                   'mariadb-server',
                    'pkg::tomcat':                  'tomcat',  # do not inherit
                    #---
                    'sys::init-sys':                'systemd',
                }),
                ((7, 1), {
                    'pacemaker[+coro]':              (1, 1, 12),
                    'pcs':                           (0, 9, 137),
                     # https://bugzilla.redhat.com/1158500
                    'resource-agents[+docker]':      (3, 9, 5),
                }),
                ((7, 2), {
                    'pacemaker[+coro]':              (1, 1, 13),
                    'pcs':                           (0, 9, 143),
                }),
                ((7, 3), {
                    'corosync':                      (2, 4),
                    'pacemaker[+coro]':              (1, 1, 15),
                    'pcs':                           (0, 9, 153),  # 152+patches
                }),
                ((7, 4), {
                    'pacemaker[+coro,+bundle]':      (1, 1, 16),
                    'pcs':                           (0, 9, 158),
                }),
                # coming...
                ((7, 5), {
                    'corosync':                      (2, 4, 3),    # guess
                    'pacemaker[+coro]':              (1, 1, 18),   # guess
                    'pcs':                           (0, 9, 162),  # guess

                }),
            ),
            'ubuntu': (
                ((13, 4), {
                    # https://packages.ubuntu.com/raring/$PACKAGE
                    'corosync':                      (1, 4),
                    'pacemaker[+coro,+hb]':          (1, 1, 7),
                    'resource-agents':               (3, 9, 2),
                    #---
                    'sys::init-sys':                'upstart',
                }),
                ((13, 10), {
                    # https://packages.ubuntu.com/saucy/$PACKAGE
                    'corosync':                      (2, 3),
                    'pacemaker[+coro,+hb]':          (1, 1, 10),
                }),
                ((14, 4), {
                    # https://packages.ubuntu.com/trusty/$PACKAGE
                    'resource-agents':               (3, 9, 3),
                }),
                ((15, 4), {
                    # https://packages.ubuntu.com/vivid/$PACKAGE
                    'pacemaker[+coro,+hb]':          (1, 1, 12),
                    #---
                    'sys::init-sys':                'systemd',
                }),
                ((16, 4), {
                    # https://packages.ubuntu.com/xenial/$PACKAGE
                    'pacemaker[+coro]':              (1, 1, 14),
                    'pcs':                           (0, 9, 149),  # universe
                    'resource-agents':               (3, 9, 7),
                }),
                ((16, 10), {
                    # https://packages.ubuntu.com/yakkety/$PACKAGE
                    'pacemaker[+coro]':              (1, 1, 15),
                    'pcs':                           (0, 9, 153),  # universe
                }),
                ((17, 4), {
                    # https://packages.ubuntu.com/zesty/$PACKAGE
                    'corosync':                      (2, 4),
                    'pacemaker[+coro]':              (1, 1, 16),
                    'pcs':                           (0, 9, 155),  # universe
                    'resource-agents':               (4, 0, 0),
                }),
                ((17, 10), {
                    # https://packages.ubuntu.com/artful/$PACKAGE
                    'pcs':                           (0, 9, 159),  # universe
                    'resource-agents':               (4, 0, 1),
                }),
            ),
        },
    'bsd':
        {
            'freebsd': (
                ((10, 4), {
                    # https://svnweb.freebsd.org/ports/tags/RELEASE_10_4_0/net/$PACKAGE
                    'corosync':                      (2, 4, 2),
                    'pacemaker[+coro]':              (1, 1, 16),
                    'resource-agents':               (4, 0, 1),
                    #---
                    'sys::init-sys':                'bsdinit',
                }),
                ((11, 0), {
                    # https://svnweb.freebsd.org/ports/tags/RELEASE_11_0_0/net/$PACKAGE
                    'corosync':                      (2, 3, 5),
                    'pacemaker[+coro]':              (1, 1, 14),
                    'resource-agents':               (3, 9, 7),
                }),
                ((11, 1), {
                    # https://svnweb.freebsd.org/ports/tags/RELEASE_11_1_0/net/$PACKAGE
                    'corosync':                      (2, 4, 2),
                    'pacemaker[+coro]':              (1, 1, 16),
                    'resource-agents':               (4, 0, 1),
                }),
            ),
        },
}

def supported_dists(sys):
    return cluster_map.get(sys, {}).keys()

# mere aliases of the distributions (packages remain the same),
# i.e., downstream rebuilders;
# values (and keys when making "alias" association) in this dict should
# correspond to `platform.linux_distribution(full_distribution_name=0)` output
# and the dict can contain also associated keys obtained as
# `platform.linux_distribution(full_distribution_name=1).lower()`
# or, on top of previous, what's expected to be entered by the user
aliases_dist = {
    # aliases
    # XXX platform.linux_distribution(full_distribution_name=0), i.e.,
    #     short distro names of Scientific Linux et al. needed
    'centos': 'redhat',
    # full_distribution_name=1 (lower-cased) -> full_distribution_name=0
    'red hat enterprise linux server': 'redhat',
    # convenience/choice of common sense or intuition
    'rhel': 'redhat',
}

# in the queries, one can use following aliases to wildcard versions
# of particular packages
aliases_rel = {
    'corosync': {
        'flatiron':   '1',
        'needle':     '2',
    },
    'debian': {  # because of http://bugs.python.org/issue9514 @ 2.6 ?
        'squeeze':      '6',
        'wheezy':       '7',
        'jessie':       '8',
        'stretch':      '9',
        'buster':      '10',
        #'buster/sid':  '10.999',  # XXX ?
        #'bullseye':    '11',
    },
    'fedora': {
        '28':         '27.999',
        'rawhide':    '27.999',
    },
    'ubuntu': {
        '13.04':      '13.4',
        'raring':     '13.4',   # Raring Ringtail
        'saucy':      '13.10',  # Saucy Salamander
        '14.04':      '14.4',
        'trusty':     '14.4',   # Trusty Tahr
        'utopic':     '14.10',  # Utopic Unicorn
        '15.04':      '15.4',
        'vivid':      '15.4',   # Vivid Vervet
        'wily':       '15.10',  # Wily Werewolf
        '16.04':      '16.4',
        'xenial':     '16.4',   # Xenial Xerus
        'yakkety':    '16.10',  # Yakkety Yak
        '17.04':      '17.4',
        'zesty':      '17.4',   # Zesty Zapus
        'artful':     '17.10',  # Artful Aardvark
        # https://wiki.ubuntu.com/BionicBeaver/ReleaseSchedule
        #'bionic':     '18.4',
        #'18.04':      '18.4',   # Bionic Beaver
    }
}

versions_extra = {
    '__cache__': {},
    'corosync': (
        ((2, 4),
            '+qdevice,+qnet'),
        ((2, 4, 3),
            '+qdevice-heuristics'),
        # tentative
        #((3, ),
        #    '+kronosnet'),
    ),
    'pacemaker': (
        # see also http://wiki.clusterlabs.org/wiki/ReleaseCalendar
        ((1, 1, 8),
            '+schema-1.2'),
        ((1, 1, 12),
            '+schema-2.0'),  # implies new style ACLs
        ((1, 1, 13),
            '+schema-2.3'),
        ((1, 1, 14),
            '+schema-2.4'),
        ((1, 1, 15),
            '+alerts,+schema-2.5'),
        ((1, 1, 16),
            '+schema-2.6'),
        ((1, 1, 17),
            # note that for our purposes, bundle ~ bundle-extra
            '+bundle,+schema-2.9'),
        ((1, 1, 18),
            '+schema-2.10'),
        # tentative
        # corosync is now required unconditionally
        #((2, 0, 0),
        #    '+corosync,+schema-3.0'),
    ),
    'pcs': (
        ((0, 9, 123),
            '+acls'),
        ((0, 9, 145),
            '+node-maintenance'),
        ((0, 9, 148),
            '+utilization'),  # https://bugzilla.redhat.com/1158500
        ((0, 9, 150),
            '+wait-cluster-start'),
        ((0, 9, 153),
            # qdevice+qnet preliminary support since 0.9.151, but...
            '+alerts,+qdevice,+qnet'),
        ((0, 9, 155),
            # this also implies support for external/* stonith agents
            '+agents-via-pacemaker'),  # https://github.com/ClusterLabs/pcs/issues/81
        # http://lists.clusterlabs.org/pipermail/users/2017-February/005103.html
        ((0, 9, 156),
            # this is best accompanied with following Pacemaker/crm_diff bugfix
            # https://github.com/ClusterLabs/pacemaker/commit/20a74b9d37
            '+push-diff'),
        # http://lists.clusterlabs.org/pipermail/users/2017-May/005824.html
        ((0, 9, 158),
            # https://bugzilla.redhat.com/1433016
            # and https://bugzilla.redhat.com/1165821, respectively
            '+bundle,+corosync-encryption-forced'),
        # http://oss.clusterlabs.org/pipermail/users/2017-June/005965.html
        ((0, 9, 159),
            # https://bugzilla.redhat.com/1447910
            # and https://bugzilla.redhat.com/1165821, respectively (1+2)
            '+bundle-meta,-corosync-encryption-forced,+corosync-encryption'),
        # http://oss.clusterlabs.org/pipermail/users/2017-October/006590.html
        #((0, 9, 160),
        #    ),
        # http://oss.clusterlabs.org/pipermail/users/2017-November/006744.html
        #((0, 9, 161),
        #    ),
        # https://oss.clusterlabs.org/pipermail/users/2018-February/014530.html
        ((0, 9, 162),
            '+qdevice-heuristics'),
        #((0, 9, 163),
        #    ),
    ),
    'resource-agents': (
        # http://lists.linux-ha.org/pipermail/linux-ha/2011-June/043321.html
        #((3, 9, 1),
        #    ),
        # http://lists.linux-ha.org/pipermail/linux-ha/2011-June/043416.html
        #((3, 9, 2),
        #    ),
        # http://lists.linux-ha.org/pipermail/linux-ha/2012-May/045193.html
        ((3, 9, 3),
            '+named'),
        # http://lists.linux-ha.org/pipermail/linux-ha/2012-November/046049.html
        ((3, 9, 4),
            '+IPaddr2~IPv6'),
        # http://lists.linux-ha.org/pipermail/linux-ha/2013-February/046461.html
        #((3, 9, 5),
        #    ),
        # http://lists.linux-ha.org/pipermail/linux-ha/2015-January/048523.html
        # XXX clvm
        ((3, 9, 6),
            '+docker'),
        # http://oss.clusterlabs.org/pipermail/users/2016-February/002216.html
        #((3, 9, 7),
        #    ),
        # http://oss.clusterlabs.org/pipermail/users/2017-January/004919.html
        #((4, 0, 0),
        #    ),
        # http://oss.clusterlabs.org/pipermail/users/2017-February/004957.html
        #((4, 0, 1),
        #    ),
        # http://oss.clusterlabs.org/pipermail/users/2017-November/006871.html
        #((4, 1, 0),
        #    ),
        # https://oss.clusterlabs.org/pipermail/users/2018-March/014578.html
        #((4, 1, 1),
        #    ),
    ),
}

#
# ...then, define some executable inference rules, starting with their helpers:
#

def _parse_ver(s):
    name, ver = (lambda a, b=None: (a, b))(*s.split('=', 1))
    if ver:
        try:
            ver = aliases_rel[name][ver]
            log.debug("resolved: {0}".format(ver))
        except KeyError:
            pass
        ver = tuple(map(int, ver.split('.')))
    return name, ver


def _cmp_ver(v1, v2, asymmetric=False):
    """Compare two given encodings of the version

    Parameters:
        v1            first operand (version tuple if not None)
        v2            second operand (version tuple if not None)
        asymmetric    do not compare absolutely, just the prefix match
    """
    if v1 and v2:
        len_v1, len_v2 = len(v1), len(v2)
        len_shorter = len_v1 if len_v1 <= len_v2 else len_v2
        for i, i1, i2 in enumerate_u(zip_longest(v1, v2, fillvalue=0)):
            if i1 == i2:
                continue
            if asymmetric and i > len_shorter:
                break
            return 1 if i1 > i2 else -1
    return 0


def _parse_extra(s, version=None, version_map=versions_extra):
    # Formalized "extra" specification:
    # EXTRA_SPEC(MODS) ::= [ EXTRAS(MODS) ]
    # EXTRAS(MODS)     ::= EXTRA(MODS) | EXTRA(MODS), EXTRAS(MODS)
    # EXTRA(MODS)      ::=(for some modifier in MODS) modifier word
    #
    # depending on the context of use, MODS can be:
    # - '' (empty string): parsing "extra" for public use, i.e.,
    #   when `version` and `version_map` not provided
    # - ''/'+'/'-': parsing "extra" for private use, i.e., when those
    #   parameters are provided
    #
    # `version_map` is a mapping from component to an ordered pairs
    # (VERSION, EXTRA_SPECS(('+', '-'))), where VERSION is encoded as
    # a tuple of numerical components of the version (i.e., result of
    # `_parse_ver`) and '+'/'-' modifier denotes the extras/features
    # being introduced/removed as of a respective version of the
    # component; `version` is a search entry between respective
    # versions here and the intermediate result of this search
    # is an extras snapshot (subsuming all the +/- changes till
    # that moment) as of the closest lower version than `version`.
    #
    # Virtually, the parsed extras are, as a sequence, appended after
    # the snapshot obtained per previous paragraph, and following rules
    # are applied on the resulting ordered sequence:
    #
    #     |- +foo [,...]  ->  foo |- [,...]
    #     |- -foo [,...]  ->      |- [,...]  # warning emitted
    #
    # foo |- +foo [,...]  ->  foo |- [,...]  # idempotency
    # foo |- -foo [,...]  ->      |- [,...]  # elimination
    name, extra = (lambda a, b='': (a, b.rstrip(']')))(*s.split('[', 1))
    allowed_mods, ret, worklist = ' ', set(), []
    if version and version_map:
        allowed_mods = '+-'
        try:
            ret = version_map['__cache__'][name][version]
        except KeyError:
            worklist = list(version_map.get(name, ()))
    else:
        version, version_map = (), {}
    worklist.append((version, extra))
    worklist.reverse()

    while worklist:
        v, es = worklist.pop()
        c = 0
        if worklist:
            c = _cmp_ver(version, v)
        elif version_map:
            # cache the resolved "snapshot" for name and version
            per_comp = version_map['__cache__'].setdefault(name, {})
            per_comp[version] = ret.copy()
        if c >= 0:
            for e in filter(len, (e.strip() for e in es.split(','))):
                if not version:
                    mod = allowed_mods
                else:
                    mod, e = (lambda a, *b: (a, ''.join(b)))(*e)
                if e and mod in allowed_mods:
                    try:
                        (set.remove if mod == '-' else set.add)(ret, e)
                    except KeyError:
                        log.warn("unexpected use of `{0}{1}'".format(mod, e))
                else:
                    log.warn("unexpected use of `{0}{1}',"
                             " unexpected/missing modifier".format(mod, e))
        else:
            worklist[:] = worklist[0:1]

    return name, ret


def infer_error(smth, branches):
    raise RuntimeError("This should not be called")


def infer_sys(sys, branches=None):
    log.debug("infer_sys: {0}: {1}".format(sys, branches))
    # lists of system-level dicts
    # -> list of dist-level dicts pertaining the specified system(s)
    if branches is None:
        branches = [cluster_map]
    if sys == "*":
        return apply_intercalate([list(iter_values(b)) for b in branches])
    return [b[sys] for b in branches if sys in b]


def infer_dist(dist, branches=None):
    # list of dist-level dicts
    # -> list of component-level dicts pertaining the specified dist(s)
    # incl. dist alias resolution
    log.debug("infer_dist: {0}: {1}".format(dist, branches))
    if branches is None:
        branches = infer_sys('*')  # alt.: branches = [cluster_map.values()]
    if dist == '*':
        return apply_intercalate([per_distver[1] for b in branches
                                  for per_dist in iter_values(b)
                                  for per_distver in per_dist])
    ret = []
    dist, dist_ver = _parse_ver(dist)
    try:
        dist = aliases_dist[dist]
    except KeyError:
        pass
    for b in branches:
        for d in b:
            if d == dist:
                # first time, we (also) traverse whole sequence of per-distro
                # releases, in-situ de-sparsifying particular packages releases;
                # to avoid needlessly repeated de-sparsifying, we are using
                # '__proceeded__' mark to denote already proceeded dicts
                d_branches = b[d]
                cur_acc = {}
                if '__proceeded__' not in d_branches or dist_ver:
                    for i, (dver, dver_branches) in enumerate(d_branches):
                        if dist_ver:
                            if (_cmp_ver(dist_ver, dver) >= 0 and
                                (i == len(d_branches) - 1
                                or _cmp_ver(dist_ver, d_branches[i+1][0]) < 0)):
                                ret.append(dver_branches)
                                if '__proceeded__' in dver_branches:
                                    break
                                else:
                                    dist_ver = None  # only desparsify since now
                        if '__proceeded__' in dver_branches:
                            continue  # only searching, no hit yet

                        for k, v in iter_items(dver_branches):
                            kk, k_extra = _parse_extra(k, version=v,
                                                       version_map=versions_extra)

                            # prevent leaking of extra from previous cycles
                            prev_extra = cur_acc.get(kk, '')
                            cur_acc.pop("{0}{1}".format(kk, prev_extra), None)

                            k_extra = ','.join(sorted(k_extra)).join("[]" if
                                                                     k_extra
                                                                     else '')
                            cur_acc[kk] = k_extra
                            cur_acc["{0}{1}".format(kk, k_extra)] = v
                        dver_branches.clear()
                        dver_branches.update(cur_acc)
                        dver_branches['__proceeded__'] = '[true]'

                if dist_ver is None and not ret:  # alt. above: dist_ver = ''
                    ret.extend(dver_branches for _, dver_branches in d_branches)
    return ret


def infer_comp(comp, branches=None):
    log.debug("infer_comp: {0}: {1}".format(comp, branches))
    # list of component-level dicts
    # -> list of component-level dicts pertaining the specified comp(s)
    # incl. component version alias resolution
    if branches is None:
        branches = infer_dist('*')  # alt.: branches = [cluster_map.values()]
    if comp == '*':
        return branches
    ret = []
    comp, comp_ver = _parse_ver(comp)
    comp, comp_extra = _parse_extra(comp)
    for b in branches:
        for c, c_ver in iter_items(b):
            c, c_extra = _parse_extra(c)
            if c == comp:
                if (isinstance(c_ver, tuple)
                        and _cmp_ver(comp_ver, c_ver, asymmetric=True) > 0
                        or comp_extra and comp_extra.difference(c_extra)):
                    continue
                ret.append(b)
                break

    return ret

rule_error = (0, infer_error)
inference_rules = {
    # type (of clause): (handling priority, handler)
    'error': rule_error,
    'sys':   (1, infer_sys),
    'dist':  (2, infer_dist),
    'comp':  (3, infer_comp),
}


#
# ...and application-specific inference engine:
#

def infer(query, system=None, system_extra=None):
    """Get/infer answers for system-distro-release-package-version queries

    Query grammar is (currently = least resistance, generalization needed):

    QUERY      ::= TERM | TERM WS* '+' WS* QUERY
    TERM       ::= TYPE:SPEC | comp:COMPSPEC  # comp~component
    WS         ::= [ ]                        # whitespace
    TYPE       ::= sys | dist                 # sys~system, dist~distro
    SPEC       ::= NAME | NAME=SUBSPEC
    COMPSPEC   ::= NAME | NAME-EXTRA | NAME=COMPSUBSPEC | NAME-EXTRA=COMPSUBSPEC
    NAME       ::= [a-zA-Z_-]+                # generally anything reasonable
    SUBSPEC    ::= NUMBER '.' NUMBER | '*'    # arbitrary version as such
    NAME-EXTRA ::= NAME '[' EXTRAS ']'
    EXTRAS     ::= NAME | NAME ',' EXTRAS
    COMPSUBSPEC::= NUMBER '.' NUMBER | NUMBER ',' '*'  # arbitrary minor version
    NUMBER     ::= [0-9]+

    with notes:

    - (COMP)SUBSPEC can be defined as a single number, minor version is assumed
      to be 0 in that case, but it's more like a syntactic sugar thant the
      full-fledged grammar case

    - for simplier expressions, several alias resolutions are in place:
        . distro (almost same set of packages known under different names)
        . component version codename (usually to denote whole major releases)

    - for working examples, see the define predicates below

    """
    # XXX trace_back=comp/dist/sys to anchor the result back in the higher sets
    # XXX only AND is supported via '+', also OR ('/'? even though in some uses
    #     '+' denotes this) together with priority control (parentheses) would
    #     be cool :)
    prev, ret = None, None
    q = [p.strip().split(':', 1) for p in query.split('+')]
    if system:
        q.append(('sys', system.lower()))
    if system_extra:
        q.append(('dist', '='.join(system_extra[:2]).lower()))
    q.sort(key=lambda t: inference_rules.get(t[0], rule_error)[0])
    level = ''
    for q_type, q_content in q:
        inference_rule = inference_rules.get(q_type, rule_error)[1]
        if q_type != level:
            prev = ret
        if not q_content:
            raise FactsError("Cannot resolve `{0}' when empty".format(q_type))
        inferred = inference_rule(q_content, prev)
        log.debug("inferred: {0}".format(inferred))
        ret = [i for i in ret if i in inferred] if q_type == level else inferred
        level = q_type
    return ret


#
# and finally, some wrapping predicates:
#

def cluster_pcs_flatiron(*sys_id):
    """Whether particular `sys_id` corresponds to pcs-flatiron cluster system"""
    return bool(infer("comp:corosync=flatiron + comp:pacemaker[cman]", *sys_id))


def cluster_pcs_needle(*sys_id):
    """Whether particular `sys_id` corresponds to pcs-needle cluster system"""
    return bool(infer("comp:corosync=needle + comp:pacemaker[coro]", *sys_id))


def cluster_pcs_1_2(*sys_id):
    """Whether particular `sys_id` corresponds to pacemaker with 1.2+ schema"""
    return bool(infer("comp:pacemaker[schema-1.2]", *sys_id))


def _find_meta(meta, which, *sys_id, **kwargs):
    meta_comp = '::'.join((meta, which))
    res = infer(':'.join(('comp', meta_comp)), *sys_id)
    for i in res:
        if meta_comp in i:
            return i[meta_comp]
    else:
        return kwargs.get('default')


def package(which, *sys_id):
    default, _ = _parse_extra(which)
    return _find_meta('pkg', which, *sys_id, default=default)


def component_or_state(which, *sys_id, **kwargs):
    """Return empty string if satisfying component found, diagnostics if not"""
    ret, found = '', infer(':'.join(('comp', which)), *sys_id)
    if not found:
        pkg, extra = _parse_extra(which)
        system, system_extra = sys_id
        ret_acc, sysextra = [], list(system_extra[:2])
        while sysextra:
            found = infer(':'.join(('comp', pkg)), system, sysextra)
            for i in found:
                for j in i:
                    p, e = _parse_extra(j)
                    if p == pkg:
                        ret_item = ''
                        if len(sysextra) == 2:
                            ret_item = p + '='.join(
                                (', '.join(e).join('[]' if e else ''),
                                 '.'.join(str(x) for x in i[j]) if not
                                 isinstance(i[j], basestring) else i[j])
                            )
                        else:
                            ret_item = kwargs.get(
                                'notyetmsg', '({0} yet untracked)'
                            ).format(p)
                        ret_acc.append(ret_item)
                        if len(sysextra) == 1:
                            break
                else:
                    continue
                break
            if ret_acc:
                break
            sysextra.pop()

        if not ret_acc:
            ret = kwargs.get('nohitmsg', '(no hit for {0})').format(pkg)
        else:
            ret = ' / '.join((', '.join(ret_acc),
                              '-'.join(args2sgpl(system, *system_extra))))

    return ret


def system(which, *sys_id):
    return _find_meta('sys', which, *sys_id, default=which)


def cmd_pkg_install(pkgs, *sys_id):
    # ready to deal with pkgs being a generator
    cmd = _find_meta('cmd', 'pkg-install', *sys_id)
    packages = ' '.join(pkgs)
    return cmd.format(packages=packages) if cmd and packages else ''


cluster_systems = (cluster_pcs_flatiron, cluster_pcs_needle)


def cluster_unknown(*sys_id):
    return not(any(cluster_sys(*sys_id) for cluster_sys in cluster_systems))


def format_dists(verbosity=0, sys='', aliases_dist_inv={}, aliases_rel_inv={}):
    # need to flip the translation tables first (if not already)
    if not aliases_dist_inv:
        aliases_dist_inv.update(reduce_u(
            lambda acc, k, v: (acc.setdefault(v, []).append(k), acc)[1],
            iter_items(aliases_dist), {}
        ))
    if not aliases_rel_inv:
        aliases_rel_inv.update((
            k,
            reduce_u(
                lambda a, ik, iv: (a.setdefault(iv, []).append(ik), a)[1],
                iter_items(v), {}
            )
        ) for k, v in iter_items(aliases_rel) if k in supported_dists(sys))

    return '\n'.join(chain(*reduce(add, (((system.join(('### ', ' ###')), ),
        ('\n\t'.join(args2sgpl(
            '\t# aliases: '.join(
                args2sgpl(k, *filter(
                    len, ('|'.join(sorted(aliases_dist_inv.get(k, ()))), )
                ))
            ),
            *tuple(
                '\t# aliases: '.join(
                    args2sgpl(
                        vvv,
                        *filter(
                            len,
                            ('|'.join(sorted(
                                aliases_rel_inv.get(k, {}) .get(vvv, ())
                            )), )
                        )
                    )
                ) for vv in (v if verbosity else (v[0], ('..', ), v[-1]))
                for vvv in ('.'.join(str(i) for i in vv[0]), )
            )
        )) for k, v in sorted(iter_items(system_map)))
    ) for system, system_map in sorted(iter_items(cluster_map))
        if verbosity or not(sys) or system == sys
    ))))



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....utils_xslt import xslt_is_member

cib2pcscmd_contopts_whitelist = (
    'image',
    'replicas',
    'replicas_per_host',
    'masters',
    'run-command',
    'network',
    'options',
)

cib2pcscmd = ('''\
    <xsl:when test="name() = 'docker'">
        <xsl:value-of select="name()"/>

        <xsl:for-each select="@*">
            <xsl:choose>
                <xsl:when test="
''' + (
                xslt_is_member('name()', cib2pcscmd_contopts_whitelist)
) + '''">
                    <xsl:value-of select='concat(" &apos;", name(), "=", ., "&apos;")'/>
                </xsl:when>
                <xsl:otherwise>
                    <xsl:message>
                        <xsl:value-of select="concat('WARNING: dropping non-whitelisted',
                                                     ' bundle/docker property: `',
                                                     name(), '`')"/>
                    </xsl:message>
                </xsl:otherwise>
            </xsl:choose>
        </xsl:for-each>
    </xsl:when>
''')



# -*- coding: UTF-8 -*-
# Copyright 2016 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....utils_xslt import xslt_is_member

###

ccs_disable_rg = '''\
    <xsl:copy>
        <xsl:attribute name="disabled">1</xsl:attribute>
    </xsl:copy>
'''

###

ccs_obfuscate_identifiers = '''\

    <!-- FAILOVER-DOMAIN -->

    <xsl:variable name="FailoverDomain"
                  select="//rm/failoverdomains/failoverdomain[@name]"/>
    <xsl:template match="
        //rm/failoverdomains/failoverdomain/@name
        |//rm/service/@domain
        |//rm/vm/@domain">
        <xsl:variable name="FailoverDomainMatch"
                      select="$FailoverDomain[
                                  @name
                                  =
                                  current()
                              ][1]"/>
        <xsl:attribute name="{name()}">
            <xsl:choose>
                <xsl:when test="$FailoverDomainMatch">
                    <!-- 1+ match(es) found -->
                    <xsl:value-of select="concat(
                        'FAILOVER-DOMAIN-',
                        count($FailoverDomainMatch/preceding-sibling::failoverdomain) + 1
                    )"/>
                </xsl:when>
                <xsl:otherwise>
                    <!-- probably refential integrity error -->
                    <xsl:value-of select="concat(
                        'FAILOVER-DOMAIN-UNDEF-',
                        generate-id()
                    )"/>
                </xsl:otherwise>
            </xsl:choose>
        </xsl:attribute>
    </xsl:template>

    <!-- SERVICE -->

    <xsl:variable name="Service"
                  select="//rm/service[@name or @ref]"/>
    <xsl:template match="
        //rm/service/@name
        |//rm/service/@ref
        |//rm/resource/@name">
        <xsl:variable name="ServiceMatch"
                      select="$Service[
                                  @name and @name = current()
                                  or
                                  @ref and @ref = current()
                              ][1]"/>
        <xsl:attribute name="{name()}">
            <xsl:choose>
                <xsl:when test="$ServiceMatch">
                    <!-- 1+ match(es) found -->
                    <xsl:value-of select="concat(
                        'SERVICE-',
                        count($ServiceMatch/preceding-sibling::service) + 1
                    )"/>
                </xsl:when>
                <xsl:otherwise>
                    <!-- unused service or refential integrity error -->
                    <xsl:value-of select="concat(
                        'SERVICE-UNUSED-',
                        generate-id()
                    )"/>
                </xsl:otherwise>
            </xsl:choose>
        </xsl:attribute>
    </xsl:template>
'''

###

from ....utils_cib import ResourceSpec
from ....utils_xslt import xslt_id_friendly
from .... import package_name

ccsflat2cibprelude_elems_res_hybrid = (
    'vm',
)

ccsflat2cibprelude_elems_res_toplevel = ccsflat2cibprelude_elems_res_hybrid + (
    'service',
)

ccsflat2cibprelude = ('''\
    <xsl:for-each select="*[
''' + (
        xslt_is_member('name()', ccsflat2cibprelude_elems_res_hybrid)
) + ''']
    |
    *[
''' + (
        xslt_is_member('name()', ccsflat2cibprelude_elems_res_toplevel)
) + ''']/*[name() != 'action']">
        <!-- meta-primary can be, e.g., @address in case of ip,
             and that can contain '/' which is not NCNameChar -->
        <xsl:variable name="Spec"
                      select="
''' + (
                              xslt_id_friendly(
                                  '(@*[name() = current()/@rgmanager-meta-primary]'
                                  '|@name'
                                  '|@address'
                                  '|@SID'
                                  '|@InstanceName'
                                  ')[1]'
                              )
) + '''"/>
        <xsl:variable name="Prefix"
                      select="concat('RESOURCE-', name(), '-', $Spec)"/>
        <primitive id="{$Prefix}">

            <xsl:attribute name="description"
            ><xsl:value-of select="concat('natively converted from ', name(),
                                          ' RA')"
            /></xsl:attribute>

            <xsl:choose>

                <!-- make sure there is at least a single "when" -->
                <xsl:when test="false()"/>

                <!-- the nested snippets should be guarded with
                     xsl:when test="name() = 'foo'" -->
                <clufter:descent-mix at="*"/>

                <xsl:otherwise>

                    <xsl:attribute name="description"
                    ><xsl:value-of select="concat('could not natively convert',
                                                  ' from ', name(), ' RA')"
                    /></xsl:attribute>
''' + (
                    ResourceSpec('ocf:rgmanager:PLACEHOLDER').xsl_attrs
) + '''
                    <xsl:attribute name="type"
                    ><xsl:value-of select="substring-before(
                                               @rgmanager-meta-agent,
                                               '.metadata'
                                           )"
                    /></xsl:attribute>
                    <instance_attributes id="{$Prefix}-ATTRS">
                        <xsl:for-each select="@*[
                                                  name() != 'rgmanager-meta-agent'
                                              ]">
                            <nvpair id="{$Prefix}-ATTRS-{name()}"
                                    name="{name()}"
                                    value="{.}"/>
                        </xsl:for-each>
                    </instance_attributes>
                    <xsl:comment> %(note_unhandled)s </xsl:comment>
                    <xsl:message>%(note_unhandled)s</xsl:message>
                </xsl:otherwise>
            </xsl:choose>

            <xsl:if test="number(@__max_failures) &gt; 0
                          and
                          number(@__failure_expire_time) &gt; 0">
                <xsl:message>
                    <xsl:value-of select="concat('WARNING: failures in status',
                                                 ' checks (akin to __max_failures',
                                                 ' and __failure_expire_time) are',
                                                 ' always directly propagated',
                                                 ' in Pacemaker without any',
                                                 ' thresholding (', $Spec, ')')"/>
                </xsl:message>
            </xsl:if>

            <!-- store service reference and auxiliary values for later use -->

            <meta_attributes id="{$Prefix}-META">
                <xsl:if test="
''' + (
                    xslt_is_member('name(..)',
                                   ccsflat2cibprelude_elems_res_toplevel)
) + '''">
                    <nvpair id="{$Prefix}-META-service"
                            name="rgmanager-service"
                            value="{concat(
                                    translate(
                                        name(..),
                                        'abcdefghijklmnopqrstuvwxyz',
                                        'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
                                    ),
                                    '-',
                                    ../@name
                                )}"/>
                </xsl:if>
                <xsl:variable name="IntervalFound"
                              select="action[
                                  (
                                      @name = 'monitor'
                                      or
                                      @name = 'status'
                                  )
                                  and
                                  not(preceding-sibling::action[
                                      (
                                          @name = 'monitor'
                                          or
                                          @name = 'status'
                                      )
                                      and
                                      @depth &lt; current()/@depth
                                  ])
                                  and
                                  not(following-sibling::action[
                                      (
                                          @name = 'monitor'
                                          or
                                          @name = 'status'
                                      )
                                      and
                                      @depth &lt; current()/@depth
                                  ])
                              ]/@interval"/>
                <xsl:variable name="Interval">
                    <xsl:value-of select="$IntervalFound"/>
                    <xsl:if test="not($IntervalFound)">
                        <xsl:value-of select="'60'"/>
                    </xsl:if>
                </xsl:variable>
                <nvpair id="{$Prefix}-META-monitor"
                        name="rgmanager-monitor"
                        value="{concat($Interval, 's')}"/>
                <!--nvpair id="{$Prefix}-META-domain"
                        name="rgmanager-domain"
                        value="{../@domain}"/-->

                <xsl:variable name="IndependentSubtree">
                    <xsl:choose>
                        <xsl:when test="number(@__independent_subtree) = 1">
                            <!-- xsl:value-of select="'1'"/ -->
                            <xsl:message>
                                <xsl:value-of select="concat('WARNING: __independent_subtree=1',
                                                            ' property not propagated as there',
                                                            ' is no straightforward equivalent',
                                                            ' (', $Spec, ')')"/>
                            </xsl:message>
                        </xsl:when>
                        <xsl:when test="number(@__independent_subtree) = 2">
                            <xsl:value-of select="'2'"/>
                        </xsl:when>
                    </xsl:choose>
                </xsl:variable>
                <xsl:if test="$IndependentSubtree != ''">
                    <nvpair id="{$Prefix}-META-independent"
                            name="rgmanager-independent"
                            value="{$IndependentSubtree}"/>
                </xsl:if>
            </meta_attributes>
        </primitive>
    </xsl:for-each>

    <xsl:for-each select="*[
''' + (
    xslt_is_member('name()', ccsflat2cibprelude_elems_res_toplevel)
) + ''']">
        <xsl:variable name="ResourceGroup"
                      select="concat(
                                  translate(
                                      name(),
                                      'abcdefghijklmnopqrstuvwxyz',
                                      'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
                                  ),
                                  '-',
                                  @name
                              )"/>
        <template id="{$ResourceGroup}"
                class="ocf"
                provider="%(package_name)s"
                type="temporary-service">
            <meta_attributes id="{$ResourceGroup}-META">
                <xsl:if test="@domain">
                    <nvpair id="{$ResourceGroup}-META-domain"
                            name="domain"
                            value="FAILOVERDOMAIN-{@domain}"/>
                </xsl:if>
                <nvpair id="{$ResourceGroup}-META-autostart"
                        name="autostart"
                        value="{@autostart}">
                    <xsl:if test="not(@autostart)">
                        <xsl:attribute name="value">1</xsl:attribute>
                    </xsl:if>
                </nvpair>
                <nvpair id="{$ResourceGroup}-META-exclusive"
                        name="exclusive"
                        value="{@exclusive}">
                    <xsl:if test="not(@exclusive)">
                        <xsl:attribute name="value">0</xsl:attribute>
                    </xsl:if>
                </nvpair>
                <nvpair id="{$ResourceGroup}-META-recovery"
                        name="recovery"
                        value="{@recovery}"/>
                <nvpair id="{$ResourceGroup}-META-depend"
                        name="depend"
                        value="{@depend}"/>
                <nvpair id="{$ResourceGroup}-META-depend_mode"
                        name="depend_mode"
                        value="{@depend_mode}"/>
                <nvpair id="{$ResourceGroup}-META-max_restarts"
                        name="max_restarts"
                        value="{@max_restarts}"/>
                <nvpair id="{$ResourceGroup}-META-restart_expire_time"
                        name="restart_expire_time"
                        value="{@restart_expire_time}"/>
                <nvpair id="{$ResourceGroup}-META-priority"
                        name="priority"
                        value="{@priority}"/>
            </meta_attributes>
        </template>
    </xsl:for-each>

    <clufter:descent at="failoverdomain"/>
''') % dict(
    package_name=package_name(),
    note_unhandled='''<xsl:value-of select="concat('WARNING: resource `',
                                                   name(),
                                                   '` is currently unhandled',
                                                   ' in the conversion, you',
                                                   ' are advised to copy ',
                                                   substring-before(
                                                       @rgmanager-meta-agent,
                                                       '.metadata'
                                                   ),
                                                   ' RGManager agent (incl.',
                                                   ' dependencies if any)',
                                                   ' to /usr/lib/ocf/resource.d',
                                                   '/rgmanager directory')"/>'''
)

###

ccs_revitalize = '''\
    <xsl:template match="service
                          |service//*[name() != 'action']
                          |vm
                          |vm//*[name() != 'action']
                          |resources/*">
        <xsl:choose>
            <!-- omit resources with repeated primary attribute
                 rgmanager/src/daemons/reslist.c:store_resource:attribute collision
                 (simplified a bit - limited just to sibling scope)
                 XXX: this will only work for config preprocessed with
                      ccs_flatten
             -->
            <xsl:when test="preceding-sibling::*[
                                name() = name(current())
                                and
                                current()/@rgmanager-meta-primary
                                and
                                @*[
                                    name() = current()/@rgmanager-meta-primary
                                ] = current()/@*[
                                    name() = current()/@rgmanager-meta-primary
                                ]
                             ]">
                <xsl:message>
                    <xsl:value-of select="concat('WARNING: omitting resource',
                                                ' with repeated primary',
                                                ' attribute (', name(), '/@',
                                                @rgmanager-meta-primary, '=',
                                                @*[
                                                    name()
                                                    =
                                                    current()/@rgmanager-meta-primary
                                                ], ')'
                                          )"/>
                </xsl:message>
            </xsl:when>
            <xsl:otherwise>
                <xsl:copy>
                    <xsl:choose>
                        <!-- omit __max_failures + __failure_expire_time when
                             used without any real effect -->
                        <xsl:when test="not(@__max_failures)
                                        !=
                                        not(@__failure_expire_time)
                                        or
                                        number(@__max_failures) &lt;= 0
                                        or
                                        number(@__failure_expire_time) &lt; 0">
                            <xsl:message>
                                <xsl:value-of select="concat('WARNING: __max_failures',
                                                             ' and __failure_expire_time',
                                                             ' must be specified in pair',
                                                             ' and both of them has to be',
                                                             ' positive numbers; not',
                                                             ' satisfied, hence dropped',
                                                             '(', name(),')')"/>
                            </xsl:message>
                            <xsl:apply-templates select="@*[
                                                             not(
''' + (
                                xslt_is_member('name()',
                                               ('__max_failures',
                                                '__failure_expire_time'))
) + ''')]"/>
                        </xsl:when>
                        <xsl:otherwise>
                            <xsl:apply-templates select="@*"/>
                        </xsl:otherwise>
                    </xsl:choose>
                    <xsl:apply-templates select="@*|node()"/>
                </xsl:copy>
            </xsl:otherwise>
        </xsl:choose>
    </xsl:template>
'''



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Various functional-paradigm helpers"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from collections import deque
from functools import reduce

from .utils import areinstances, args2sgpl, tuplist
from .utils_2to3 import enumerate_u, reduce_u

# Non-lazy (materialized at one point) variant of map
# (http://stackoverflow.com/a/18433519).
foreach = lambda *args: deque(map(*args), maxlen=0)

apply_preserving_depth = \
    lambda action: \
        lambda item: \
            type(item)(apply_preserving_depth(action)(i) for i in item) \
            if tuplist(item) else action(item)
apply_aggregation_preserving_depth = \
    lambda agg_fn: \
        lambda item: \
            agg_fn(type(item)(apply_aggregation_preserving_depth(agg_fn)(i)
                              for i in item)) \
            if tuplist(item) else item
apply_aggregation_preserving_passing_depth = \
    lambda agg_fn, d=0: \
        lambda item: \
            agg_fn(type(item)(
                apply_aggregation_preserving_passing_depth(agg_fn, d+1)(i)
                for i in item),
                d+1
            ) \
            if tuplist(item) else item
# name comes from Haskell
# note: always returns list even for input tuple
# note: when returning from recursion, should always observe scalar or list(?)
apply_intercalate = apply_aggregation_preserving_depth(
    lambda i:
        reduce(lambda a, b: a + (b if isinstance(b, list) else [b]),
               i, [])
)

add_item = lambda seq, i: \
    seq + (i if areinstances(seq, i) else type(seq)(args2sgpl(i)))

# something like str.partition, but repeatable and applicable to sequences
# in general (can be used as a one-level only inverse function to
# apply_aggregation_preserving_depth (or, in special case, apply_intercalate),
# which itself can loose granularity while this one provide a finer view);
# note that special care was taken to make this function
# uniformly handle both strings and generic iterables (unlike the other
# functions)
apply_partition = \
    lambda i, partitioner: \
        (lambda result: result[:-1] if result[-1] == type(i)() else result)(
            reduce_u(
                lambda acc, ee, ii:
                    acc[:-1] + [add_item(acc[-1], ii)]
                    if not partitioner(ii, ee, acc) else (acc if ee else [])
                        + [add_item(type(i)(), ii)] + [type(i)()],
                enumerate_u(i),
                [type(i)()]
            )
        )

# ...and also something like str.split (unlimited variant)
apply_split = \
    lambda i, splitter: \
        (lambda result: result[:-1] if result[-1] == type(i)() else result)(
            reduce_u(
                lambda acc, ee, ii:
                    acc[:-1] + [add_item(acc[-1], ii)]
                    if not splitter(ii, ee, acc) else acc + [type(i)()],
                enumerate_u(i),
                [type(i)()]
            )
        )

# Haskell: partition
bifilter = \
    lambda fnc, seq: \
        reduce(lambda acc, x: acc[int(not fnc(x))].append(x) or acc,
               seq, ([], []))

## + 2-to-3 unpacking version
bifilter_unpack = \
    lambda fnc, seq: \
        reduce(lambda acc, x: acc[int(not fnc(*x))].append(x) or acc,
               seq, ([], []))

# Given the partitioning function, do the recursive refinement in a way
# the parts reaching the bottom line in more steps are continually "shaked"
# towards the end (with a stable relative order, though).
# Default partitioning function just distinguishes between tuples/lists
# and other values (presumably scalars) so when passed a tuple/list encoding
# a graph (tree/forest) in a DFS manner using unrestricted nesting, it will
# return serialized BFS walk of the same.
# NOTE if there are scalars already at the start-level depth, use
#      `tail_shake_safe` (not that it provides much more guarantee :-p)
tailshake = \
    lambda src, partitioner=(lambda x: not tuplist(x)): \
        reduce(lambda a, b: a + (tailshake(b, partitioner) if b else b),
               reduce(lambda a, b: (a[0] + b[0], a[1] + b[1]),
                      tuple(bifilter(partitioner, i) if tuplist(i) else (i, [])
                      for i in src), ([], [])))

tailshake_safe = \
    lambda src, partitioner=(lambda x: not tuplist(x)): \
        tailshake(src if all(tuplist(i) for i in src) else (src, ),
                  partitioner)


zipped_outlier = type('zipped_outlier', (tuple, ), {})
zip_empty = type('zip_filler', (str, ), {})("EMPTY")
loose_zip = lambda a, b: zip(
    list(a) + (max(len(a), len(b)) - len(a)) * [zip_empty],
    list(b) + (max(len(a), len(b)) - len(b)) * [zip_empty]
)
apply_loose_zip_preserving_depth = \
    lambda a, b: \
        type(a)(
            [apply_loose_zip_preserving_depth(*p) for p in loose_zip(a, b)]
        ) if tuplist(a) == tuplist(b) == True else zipped_outlier([a, b])
# as previous, but with length checking of some sort
# NOTE: automatically shortens the longer counterpart in the pair
#       to the length of the bigger one
apply_strict_zip_preserving_depth = \
    lambda a, b: \
        type(a)(
            [apply_strict_zip_preserving_depth(*p) for p in zip(a, b)]
        ) if tuplist(a) == tuplist(b) == True and len(a) == len(b) \
        else zipped_outlier([a, b])



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Base protocol stuff (metaclass, etc.)"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from collections import MutableMapping
from logging import getLogger

from .plugin_registry import PluginRegistry
from .utils import args2sgpl, tuplist
from .utils_2to3 import MimicMeta
from .utils_prog import cli_undecor

log = getLogger(__name__)

protodict = lambda x: isinstance(x, MutableMapping) and 'passin' in x
protodictval = lambda x: args2sgpl(x['passin']) if protodict(x) else x


class protocols(PluginRegistry):
    """Protocol registry (to be used as a metaclass for filters)

    To be noted, this harness is solely optional, and only good to allow
    early discovery of the typos in the protocols and to maintain some
    in-code documentation of their usage as opposed to much further
    in the processing with some files already successfully produced, etc.
    """
    _namespace = ''  # avoid the plugins module-import namespace, unused anyway

    @classmethod
    def register(registry, pr):
        # undecor to pass the checks in probe
        return registry.probe(cli_undecor(pr),
                              pr if isinstance(pr, Protocol) else Protocol(pr))


class _Protocol(object):
    """Class intended to be (exceptionally) instantiated (enhanced string)"""
    @MimicMeta.method
    def __new__(cls, *args, **kwargs):
        ret = super(Protocol, cls).__new__(cls, *args, **kwargs)
        return protocols.register(ret)

    @MimicMeta.method
    def ensure_proto(self, value=None):
        work_val = protodictval(value)
        work_val = work_val if tuplist(work_val) else (str(self), work_val) \
                            if work_val is not None else (str(self), )
        if protodict(value):
            value['passin'] = work_val
            work_val = value
        return work_val


Protocol = MimicMeta('Protocol', protocols, _Protocol, (str, ))



# -*- coding: UTF-8 -*-
# Copyright 2016 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Wrapper around standard lxml.etree static methods"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from lxml import etree

etree_XSLT_safe = lambda _input, **kwargs: \
    etree.XSLT(_input,
               **dict(access_control=etree.XSLTAccessControl.DENY_ALL,
                      **kwargs))

etree_parser_safe_kwargs = dict(
    no_network=True,
    #resolve_entities=False,
)

etree_parser_safe = etree.XMLParser(**etree_parser_safe_kwargs)

etree_parser_safe_unblanking = etree.XMLParser(**dict(
    remove_blank_text=True,
    **etree_parser_safe_kwargs
))



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""simpleconfig-normalize filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from logging import getLogger

log = getLogger(__name__)

from ..filter import Filter
from ..utils import selfaware
from ..utils_2to3 import iter_items


@selfaware
def _simpleconfig_normalize(me, parent_section):
    # must not attempt to modify anything from parent_section in-place
    parent_tag, parent_options, parent_sections = parent_section
    parent_sections_new = []
    for s in parent_sections:
        tag, options, sections = s
        opts = OrderedDict()
        for n, v in options:
            vals = opts.setdefault(n, [])
            if v in vals:
                log.warning("omitting duplicated `({0}, {1}}' option from"
                            " normalization".format(n, v))
            else:
                vals.append(v)
        options = []
        for n, vals in iter_items(opts):
            options.append((n, vals[0]))
            vals[:] = vals[1:]
        parent_sections_new.append((tag, options, [me(s) for s in sections]))
        for i, (n, vals) in enumerate((n, vals) for n, vals in iter_items(opts)
                                      if vals):
            if not i and sections:
                log.warning("current section `{0}' needs normalization but"
                            " contains subsections (not expected)".format(tag))
            for v in vals:
                parent_sections_new.append((tag, ((n, v), ), ()))
    return (parent_tag, parent_options, parent_sections_new)


@Filter.deco('simpleconfig', 'simpleconfig-normalized')
def simpleconfig_normalize(flt_ctxt, in_obj):
    """Normalizes simpleconfig per simpleconfig-normalized format description

    (Mentioned traversal is deliberately defined as post-order here.)
    """
    struct = in_obj('struct', protect_safe=True)
    return ('struct', _simpleconfig_normalize(struct))
